# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import textwrap

import sqlalchemy as sa
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.scripts import cleanupdb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.unit.db import test_logs
from buildbot.test.util import dirs
from buildbot.test.util import misc
from buildbot.util.twisted import async_to_deferred

try:
    import lz4

    _ = lz4
    hasLz4 = True
except ImportError:
    hasLz4 = False

try:
    import zstandard

    _ = zstandard
    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

try:
    import brotli

    _ = brotli
    HAS_BROTLI = True
except ImportError:
    HAS_BROTLI = False


def mkconfig(**kwargs):
    config = {"quiet": False, "basedir": os.path.abspath('basedir'), "force": True}
    config.update(kwargs)
    return config


def write_buildbot_tac(path):
    with open(path, "w", encoding='utf-8') as f:
        f.write(
            textwrap.dedent("""
            from twisted.application import service
            application = service.Application('buildmaster')
        """)
        )


def write_master_cfg(path, db_url, extraconfig):
    with open(path, "w", encoding='utf-8') as f:
        f.write(
            textwrap.dedent(f"""
            from buildbot.plugins import *
            c = BuildmasterConfig = dict()
            c['db_url'] = {db_url!r}
            c['buildbotNetUsageData'] = None
            c['multiMaster'] = True  # don't complain for no builders
            {extraconfig}
        """)
        )


class TestCleanupDb(
    misc.StdoutAssertionsMixin, dirs.DirsMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        self.setUpDirs('basedir')
        write_buildbot_tac(os.path.join('basedir', 'buildbot.tac'))
        self.setUpStdoutAssertions()

    def createMasterCfg(self, extraconfig=""):
        write_master_cfg(os.path.join('basedir', 'master.cfg'), 'sqlite://', extraconfig)

    @async_to_deferred
    async def test_cleanup_not_basedir(self):
        res = await cleanupdb._cleanupDatabase(mkconfig(basedir='doesntexist'))
        self.assertEqual(res, 1)
        self.assertInStdout('invalid buildmaster directory')

    @async_to_deferred
    async def test_cleanup_bad_config(self):
        res = await cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
        self.assertEqual(res, 1)
        self.assertInStdout("master.cfg' does not exist")

    @async_to_deferred
    async def test_cleanup_bad_config2(self):
        self.createMasterCfg(extraconfig="++++ # syntaxerror")
        res = await cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
        self.assertEqual(res, 1)
        self.assertInStdout("encountered a SyntaxError while parsing config file:")
        # config logs an error via log.err, we must eat it or trial will
        # complain
        self.flushLoggedErrors()


class TestCleanupDbRealDb(
    misc.StdoutAssertionsMixin, dirs.DirsMixin, TestReactorMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setUpDirs('basedir')
        write_buildbot_tac(os.path.join('basedir', 'buildbot.tac'))
        self.setUpStdoutAssertions()

        self.master = yield fakemaster.make_master(
            self, wantDb=True, wantRealReactor=True, sqlite_memory=False
        )

    def createMasterCfg(self, db_url, extraconfig=""):
        write_master_cfg(os.path.join('basedir', 'master.cfg'), db_url, extraconfig)

    @async_to_deferred
    async def test_cleanup(self):
        # we reuse the fake db background data from db.logs unit tests
        await self.master.db.insert_test_data(test_logs.Tests.backgroundData)

        # insert a log with lots of redundancy
        LOGDATA = "xx\n" * 2000
        logid = await self.master.db.logs.addLog(102, "x", "x", "s")
        await self.master.db.logs.appendLog(logid, LOGDATA)

        db_url = self.master.db.configured_db_config.db_url

        # test all methods
        lengths = {}
        for mode in self.master.db.logs.COMPRESSION_MODE:
            if mode == "lz4" and not hasLz4:
                # ok.. lz4 is not installed, don't fail
                lengths["lz4"] = 40
                continue
            if mode == "zstd" and not HAS_ZSTD:
                # zstandard is not installed, don't fail
                lengths["zstd"] = 20
                continue
            if mode == "br" and not HAS_BROTLI:
                # brotli is not installed, don't fail
                lengths["br"] = 14
                continue
            # create a master.cfg with different compression method
            self.createMasterCfg(db_url, f"c['logCompressionMethod'] = '{mode}'")
            res = await cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
            self.assertEqual(res, 0)

            # make sure the compression don't change the data we can retrieve
            # via api
            res = await self.master.db.logs.getLogLines(logid, 0, 2000)
            self.assertEqual(res, LOGDATA)

            # retrieve the actual data size in db using raw sqlalchemy
            def thd(conn):
                tbl = self.master.db.model.logchunks
                q = sa.select(sa.func.sum(sa.func.length(tbl.c.content)))
                q = q.where(tbl.c.logid == logid)
                return conn.execute(q).fetchone()[0]

            lengths[mode] = await self.master.db.pool.do(thd)

        self.assertDictAlmostEqual(
            lengths,
            {
                'raw': 5999,
                'bz2': 44,
                'lz4': 40,
                'gz': 31,
                'zstd': 20,
                'br': 14,
            },
        )

    def assertDictAlmostEqual(self, d1, d2):
        # The test shows each methods return different size
        # but we still make a fuzzy comparison to resist if underlying libraries
        # improve efficiency
        self.assertEqual(len(d1), len(d2))
        for k in d2.keys():
            self.assertApproximates(d1[k], d2[k], 10)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import steps
from buildbot.db.steps import StepModel
from buildbot.db.steps import UrlModel
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util import epoch2datetime

TIME1 = 2001111
TIME2 = 2002222
TIME3 = 2003333
TIME4 = 2004444


class StepEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = steps.StepEndpoint
    resourceTypeClass = steps.Step

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Builder(id=77, name='builder77'),
            fakedb.Master(id=88),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, builderid=77, buildsetid=8822),
            fakedb.Build(
                id=30, builderid=77, number=7, masterid=88, buildrequestid=82, workerid=47
            ),
            fakedb.Step(
                id=70,
                number=0,
                name='one',
                buildid=30,
                started_at=TIME1,
                locks_acquired_at=TIME2,
                complete_at=TIME3,
                results=0,
            ),
            fakedb.Step(
                id=71,
                number=1,
                name='two',
                buildid=30,
                started_at=TIME2,
                locks_acquired_at=TIME3,
                complete_at=TIME4,
                results=2,
                urls_json='[{"name":"url","url":"http://url"}]',
            ),
            fakedb.Step(id=72, number=2, name='three', buildid=30, started_at=TIME4, hidden=True),
        ])

    @defer.inlineCallbacks
    def test_get_existing(self):
        step = yield self.callGet(('steps', 72))
        self.validateData(step)
        self.assertEqual(
            step,
            {
                'buildid': 30,
                'complete': False,
                'complete_at': None,
                'name': 'three',
                'number': 2,
                'results': None,
                'started_at': epoch2datetime(TIME4),
                "locks_acquired_at": None,
                'state_string': '',
                'stepid': 72,
                'urls': [],
                'hidden': True,
            },
        )

    @defer.inlineCallbacks
    def test_get_existing_buildid_name(self):
        step = yield self.callGet(('builds', 30, 'steps', 'two'))
        self.validateData(step)
        self.assertEqual(step['stepid'], 71)

    @defer.inlineCallbacks
    def test_get_existing_buildid_number(self):
        step = yield self.callGet(('builds', 30, 'steps', 1))
        self.validateData(step)
        self.assertEqual(step['stepid'], 71)

    @defer.inlineCallbacks
    def test_get_existing_builder_name(self):
        step = yield self.callGet(('builders', 77, 'builds', 7, 'steps', 'two'))
        self.validateData(step)
        self.assertEqual(step['stepid'], 71)

    @defer.inlineCallbacks
    def test_get_existing_buildername_name(self):
        step = yield self.callGet(('builders', 'builder77', 'builds', 7, 'steps', 'two'))
        self.validateData(step)
        self.assertEqual(step['stepid'], 71)

    @defer.inlineCallbacks
    def test_get_existing_builder_number(self):
        step = yield self.callGet(('builders', 77, 'builds', 7, 'steps', 1))
        self.validateData(step)
        self.assertEqual(step['stepid'], 71)

    @defer.inlineCallbacks
    def test_get_missing_buildername_builder_number(self):
        step = yield self.callGet(('builders', 'builder77_nope', 'builds', 7, 'steps', 1))
        self.assertEqual(step, None)

    @defer.inlineCallbacks
    def test_get_missing(self):
        step = yield self.callGet(('steps', 9999))
        self.assertEqual(step, None)


class StepsEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = steps.StepsEndpoint
    resourceTypeClass = steps.Step

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Builder(id=77, name='builder77'),
            fakedb.Master(id=88),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, builderid=77, buildsetid=8822),
            fakedb.Build(
                id=30, builderid=77, number=7, masterid=88, buildrequestid=82, workerid=47
            ),
            fakedb.Build(
                id=31, builderid=77, number=8, masterid=88, buildrequestid=82, workerid=47
            ),
            fakedb.Step(
                id=70,
                number=0,
                name='one',
                buildid=30,
                started_at=TIME1,
                locks_acquired_at=TIME2,
                complete_at=TIME3,
                results=0,
            ),
            fakedb.Step(
                id=71,
                number=1,
                name='two',
                buildid=30,
                started_at=TIME2,
                locks_acquired_at=TIME3,
                complete_at=TIME4,
                results=2,
                urls_json='[{"name":"url","url":"http://url"}]',
            ),
            fakedb.Step(id=72, number=2, name='three', buildid=30, started_at=TIME4),
            fakedb.Step(id=73, number=0, name='otherbuild', buildid=31, started_at=TIME3),
        ])

    @defer.inlineCallbacks
    def test_get_buildid(self):
        steps = yield self.callGet(('builds', 30, 'steps'))

        for step in steps:
            self.validateData(step)

        self.assertEqual([s['number'] for s in steps], [0, 1, 2])

    @defer.inlineCallbacks
    def test_get_builder(self):
        steps = yield self.callGet(('builders', 77, 'builds', 7, 'steps'))

        for step in steps:
            self.validateData(step)

        self.assertEqual([s['number'] for s in steps], [0, 1, 2])

    @defer.inlineCallbacks
    def test_get_buildername(self):
        steps = yield self.callGet(('builders', 'builder77', 'builds', 7, 'steps'))

        for step in steps:
            self.validateData(step)

        self.assertEqual([s['number'] for s in steps], [0, 1, 2])


class Step(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = steps.Step(self.master)

        yield self.master.db.insert_test_data([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Builder(id=77, name='builder77'),
            fakedb.Master(id=88),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, builderid=77, buildsetid=8822),
            fakedb.Build(
                id=10, builderid=77, number=7, masterid=88, buildrequestid=82, workerid=47
            ),
        ])

    def test_signature_addStep(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.addStep,  # fake
            self.rtype.addStep,
        )  # real
        def addStep(self, buildid, name):
            pass

    @defer.inlineCallbacks
    def test_addStep(self):
        stepid, number, name = yield self.rtype.addStep(buildid=10, name='name')
        msgBody = {
            'buildid': 10,
            'complete': False,
            'complete_at': None,
            'name': name,
            'number': number,
            'results': None,
            'started_at': None,
            "locks_acquired_at": None,
            'state_string': 'pending',
            'stepid': stepid,
            'urls': [],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'new'), msgBody),
            (('steps', str(stepid), 'new'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name=name,
                number=number,
                results=None,
                started_at=None,
                locks_acquired_at=None,
                state_string='pending',
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_fake_addStep(self):
        self.assertEqual(len((yield self.master.data.updates.addStep(buildid=10, name='ten'))), 3)

    def test_signature_startStep(self):
        @self.assertArgSpecMatches(self.master.data.updates.startStep, self.rtype.startStep)
        def addStep(self, stepid, started_at=None, locks_acquired=False):
            pass

    @defer.inlineCallbacks
    def test_startStep(self):
        self.reactor.advance(TIME1)
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name='ten', state_string='pending'
        )
        yield self.rtype.startStep(stepid=stepid)

        msgBody = {
            'buildid': 10,
            'complete': False,
            'complete_at': None,
            'name': 'ten',
            'number': 0,
            'results': None,
            'started_at': epoch2datetime(TIME1),
            "locks_acquired_at": None,
            'state_string': 'pending',
            'stepid': stepid,
            'urls': [],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'started'), msgBody),
            (('steps', str(stepid), 'started'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name='ten',
                number=0,
                results=None,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=None,
                state_string='pending',
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_startStep_no_locks(self):
        self.reactor.advance(TIME1)
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name="ten", state_string="pending"
        )
        yield self.rtype.startStep(stepid=stepid, locks_acquired=True)

        msgBody = {
            "buildid": 10,
            "complete": False,
            "complete_at": None,
            "name": "ten",
            "number": 0,
            "results": None,
            "started_at": epoch2datetime(TIME1),
            "locks_acquired_at": epoch2datetime(TIME1),
            "state_string": "pending",
            "stepid": stepid,
            "urls": [],
            "hidden": False,
        }
        self.master.mq.assertProductions([
            (("builds", "10", "steps", str(stepid), "started"), msgBody),
            (("steps", str(stepid), "started"), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name="ten",
                number=0,
                results=None,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=epoch2datetime(TIME1),
                state_string="pending",
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_startStep_acquire_locks(self):
        self.reactor.advance(TIME1)
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name='ten', state_string='pending'
        )
        yield self.rtype.startStep(stepid=stepid)
        self.reactor.advance(TIME2 - TIME1)
        self.master.mq.clearProductions()
        yield self.rtype.set_step_locks_acquired_at(stepid=stepid)

        msgBody = {
            'buildid': 10,
            'complete': False,
            'complete_at': None,
            'name': 'ten',
            'number': 0,
            'results': None,
            'started_at': epoch2datetime(TIME1),
            "locks_acquired_at": epoch2datetime(TIME2),
            'state_string': 'pending',
            'stepid': stepid,
            'urls': [],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'updated'), msgBody),
            (('steps', str(stepid), 'updated'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name='ten',
                number=0,
                results=None,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=epoch2datetime(TIME2),
                state_string='pending',
                urls=[],
                hidden=False,
            ),
        )

    def test_signature_setStepStateString(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.setStepStateString,  # fake
            self.rtype.setStepStateString,
        )  # real
        def setStepStateString(self, stepid, state_string):
            pass

    @defer.inlineCallbacks
    def test_setStepStateString(self):
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name='ten', state_string='pending'
        )
        yield self.rtype.setStepStateString(stepid=stepid, state_string='hi')

        msgBody = {
            'buildid': 10,
            'complete': False,
            'complete_at': None,
            'name': 'ten',
            'number': 0,
            'results': None,
            'started_at': None,
            "locks_acquired_at": None,
            'state_string': 'hi',
            'stepid': stepid,
            'urls': [],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'updated'), msgBody),
            (('steps', str(stepid), 'updated'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name='ten',
                number=0,
                results=None,
                started_at=None,
                locks_acquired_at=None,
                state_string='hi',
                urls=[],
                hidden=False,
            ),
        )

    def test_signature_finishStep(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.finishStep,  # fake
            self.rtype.finishStep,
        )  # real
        def finishStep(self, stepid, results, hidden):
            pass

    @defer.inlineCallbacks
    def test_finishStep(self):
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name='ten', state_string='pending'
        )
        self.reactor.advance(TIME1)
        yield self.rtype.startStep(stepid=stepid)
        yield self.rtype.set_step_locks_acquired_at(stepid=stepid)
        self.reactor.advance(TIME2 - TIME1)
        self.master.mq.clearProductions()
        yield self.rtype.finishStep(stepid=stepid, results=9, hidden=False)

        msgBody = {
            'buildid': 10,
            'complete': True,
            'complete_at': epoch2datetime(TIME2),
            'name': 'ten',
            'number': 0,
            'results': 9,
            'started_at': epoch2datetime(TIME1),
            "locks_acquired_at": epoch2datetime(TIME1),
            'state_string': 'pending',
            'stepid': stepid,
            'urls': [],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'finished'), msgBody),
            (('steps', str(stepid), 'finished'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=epoch2datetime(TIME2),
                id=stepid,
                name='ten',
                number=0,
                results=9,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=epoch2datetime(TIME1),
                state_string='pending',
                urls=[],
                hidden=False,
            ),
        )

    def test_signature_addStepURL(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.addStepURL,  # fake
            self.rtype.addStepURL,
        )  # real
        def addStepURL(self, stepid, name, url):
            pass

    @defer.inlineCallbacks
    def test_addStepURL(self):
        stepid, _, _ = yield self.master.db.steps.addStep(
            buildid=10, name='ten', state_string='pending'
        )
        yield self.rtype.addStepURL(stepid=stepid, name="foo", url="bar")

        msgBody = {
            'buildid': 10,
            'complete': False,
            'complete_at': None,
            'name': 'ten',
            'number': 0,
            'results': None,
            'started_at': None,
            "locks_acquired_at": None,
            'state_string': 'pending',
            'stepid': stepid,
            'urls': [{'name': 'foo', 'url': 'bar'}],
            'hidden': False,
        }
        self.master.mq.assertProductions([
            (('builds', '10', 'steps', str(stepid), 'updated'), msgBody),
            (('steps', str(stepid), 'updated'), msgBody),
        ])
        step = yield self.master.db.steps.getStep(stepid)
        self.assertEqual(
            step,
            StepModel(
                buildid=10,
                complete_at=None,
                id=stepid,
                name='ten',
                number=0,
                results=None,
                started_at=None,
                locks_acquired_at=None,
                state_string='pending',
                urls=[UrlModel(name='foo', url='bar')],
                hidden=False,
            ),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import io
import json
import random
import shlex
from typing import Any
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.internet import reactor

from buildbot import config
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.reporters.words import Channel
from buildbot.reporters.words import Contact
from buildbot.reporters.words import StatusBot
from buildbot.reporters.words import UsageError
from buildbot.reporters.words import WebhookResource
from buildbot.schedulers.forcesched import CollectedValidationError
from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.util import Notifier
from buildbot.util import asyncSleep
from buildbot.util import bytes2unicode
from buildbot.util import epoch2datetime
from buildbot.util import httpclientservice
from buildbot.util import service
from buildbot.util import unicode2bytes


class TelegramChannel(Channel):
    def __init__(self, bot, channel):
        assert isinstance(channel, dict), "channel must be a dict provided by Telegram API"
        super().__init__(bot, channel['id'])
        self.chat_info = channel

    @defer.inlineCallbacks
    def list_notified_events(self):
        if self.notify_events:
            notified_events = "\n".join(sorted(f"🔔 **{n}**" for n in self.notify_events))
            yield self.send(f"The following events are being notified:\n{notified_events}")
        else:
            yield self.send("🔕 No events are being notified.")


def collect_fields(fields):
    for field in fields:
        if field['fullName']:
            yield field
        if 'fields' in field:
            yield from collect_fields(field['fields'])


class TelegramContact(Contact):
    def __init__(self, user, channel=None):
        assert isinstance(user, dict), "user must be a dict provided by Telegram API"
        self.user_info = user
        super().__init__(user['id'], channel)
        self.template = None

    @property
    def chat_id(self):
        return self.channel.id

    @property
    def user_full_name(self):
        fullname = " ".join((
            self.user_info['first_name'],
            self.user_info.get('last_name', ''),
        )).strip()
        return fullname

    @property
    def user_name(self):
        return self.user_info['first_name']

    def describeUser(self):
        user = self.user_full_name
        try:
            user += f" (@{self.user_info['username']})"
        except KeyError:
            pass

        if not self.is_private_chat:
            chat_title = self.channel.chat_info.get('title')
            if chat_title:
                user += f" on '{chat_title}'"

        return user

    ACCESS_DENIED_MESSAGES = [
        "🧙‍♂️ You shall not pass! 👹",
        "😨 Oh NO! You are simply not allowed to this! 😢",
        "⛔ You cannot do this. Better go outside and relax... 🌳",
        "⛔ ACCESS DENIED! This incident has ben reported to NSA, KGB, and George Soros! 🕵",
        "🚫 Unauthorized access detected! Your device will explode in 3... 2... 1... 💣",
        "☢ Radiation level too high! Continuation of the procedure forbidden! 🛑",
    ]

    def access_denied(self, *args, tmessage, **kwargs):
        self.send(
            random.choice(self.ACCESS_DENIED_MESSAGES), reply_to_message_id=tmessage['message_id']
        )

    def query_button(self, caption, payload):
        if isinstance(payload, str) and len(payload) < 64:
            return {'text': caption, 'callback_data': payload}
        key = hash(repr(payload))
        while True:
            cached = self.bot.query_cache.get(key)
            if cached is None:
                self.bot.query_cache[key] = payload
                break
            if cached == payload:
                break
            key += 1
        return {'text': caption, 'callback_data': key}

    @defer.inlineCallbacks
    def command_START(self, args, **kwargs):
        self.bot.post(
            '/setChatMenuButton',
            json={'chat_id': self.channel.id, 'menu_button': {'type': 'commands'}},
        )
        yield self.command_HELLO(args)
        reactor.callLater(0.2, self.command_HELP, '')

    def command_NAY(self, args, tmessage, **kwargs):
        """forget the current command"""
        replied_message = tmessage.get('reply_to_message')
        if replied_message:
            if 'reply_markup' in replied_message:
                self.bot.edit_keyboard(self.channel.id, replied_message['message_id'])
        if self.is_private_chat:
            self.send("Never mind...")
        else:
            self.send(f"Never mind, {self.user_name}...")

    command_NAY.usage = "nay - never mind the command we are currently discussing"  # type: ignore[attr-defined]

    @classmethod
    def get_commands(cls):
        commands = cls.build_commands()
        response = []
        for command in commands:
            if command == 'start':
                continue
            meth = getattr(cls, 'command_' + command.upper())
            doc = getattr(meth, '__doc__', None)
            if not doc:
                doc = command
            response.append((command, doc))
        return response

    @classmethod
    def describe_commands(cls):
        return [f"{command} - {doc}" for command, doc in cls.get_commands()]

    @Contact.overrideCommand
    def command_COMMANDS(self, args, **kwargs):
        if args.lower() == 'botfather':
            response = self.describe_commands()
            if response:
                self.send('\n'.join(response))
        else:
            return super().command_COMMANDS(args)
        return None

    @defer.inlineCallbacks
    def command_GETID(self, args, **kwargs):
        """get user and chat ID"""
        if self.is_private_chat:
            self.send(f"Your ID is `{self.user_id}`.")
        else:
            yield self.send(f"{self.user_name}, your ID is `{self.user_id}`.")
            self.send(f'This {self.channel.chat_info.get("type", "group")} ID is `{self.chat_id}`.')

    command_GETID.usage = (  # type: ignore[attr-defined]
        "getid - get user and chat ID that can be put in the master configuration file"
    )

    @defer.inlineCallbacks
    @Contact.overrideCommand
    def command_LIST(self, args, **kwargs):
        args = self.splitArgs(args)
        if not args:
            keyboard = [
                [
                    self.query_button("👷️ Builders", '/list builders'),
                    self.query_button("👷️ (including old ones)", '/list all builders'),
                ],
                [
                    self.query_button("⚙ Workers", '/list workers'),
                    self.query_button("⚙ (including old ones)", '/list all workers'),
                ],
                [self.query_button("📄 Changes (last 10)", '/list changes')],
            ]
            self.send("What do you want to list?", reply_markup={'inline_keyboard': keyboard})
            return

        all = False
        num = 10
        try:
            num = int(args[0])
            del args[0]
        except ValueError:
            if args[0] == 'all':
                all = True
                del args[0]
        except IndexError:
            pass

        if not args:
            raise UsageError(
                "Try '" + self.bot.commandPrefix + "list [all|N] builders|workers|changes'."
            )

        if args[0] == 'builders':
            bdicts = yield self.bot.getAllBuilders()
            online_builderids = yield self.bot.getOnlineBuilders()

            response = ["I found the following **builders**:"]
            for bdict in bdicts:
                if bdict['builderid'] in online_builderids:
                    response.append(f"`{bdict['name']}`")
                elif all:
                    response.append(f"`{bdict['name']}` ❌")
            self.send('\n'.join(response))

        elif args[0] == 'workers':
            workers = yield self.master.data.get(('workers',))

            response = ["I found the following **workers**:"]
            for worker in workers:
                if worker['configured_on']:
                    response.append(f"`{worker['name']}`")
                    if not worker['connected_to']:
                        response[-1] += " ⚠️"
                elif all:
                    response.append(f"`{worker['name']}` ❌")
            self.send('\n'.join(response))

        elif args[0] == 'changes':
            wait_message = yield self.send("⏳ Getting your changes...")

            if all:
                changes = yield self.master.data.get(('changes',))
                self.bot.delete_message(self.channel.id, wait_message['message_id'])
                num = len(changes)
                if num > 50:
                    keyboard = [
                        [
                            self.query_button(
                                "‼ Yes, flood me with all of them!", f'/list {num} changes'
                            )
                        ],
                        [self.query_button("✅ No, just show last 50", '/list 50 changes')],
                    ]
                    self.send(
                        f"I found {num} changes. Do you really want me to list them all?",
                        reply_markup={'inline_keyboard': keyboard},
                    )
                    return

            else:
                changes = yield self.master.data.get(('changes',), order=['-changeid'], limit=num)
                self.bot.delete_message(self.channel.id, wait_message['message_id'])

            response = ["I found the following recent **changes**:\n"]

            for change in reversed(changes):
                change['comment'] = change['comments'].split('\n')[0]
                change['date'] = epoch2datetime(change['when_timestamp']).strftime('%Y-%m-%d %H:%M')
                response.append(
                    f"[{change['comment']}]({change['revlink']})\n"
                    f"_Author_: {change['author']}\n"
                    f"_Date_: {change['date']}\n"
                    f"_Repository_: {change['repository']}\n"
                    f"_Branch_: {change['branch']}\n"
                    f"_Revision_: {change['revision']}\n"
                )
            self.send('\n'.join(response))

    @defer.inlineCallbacks
    def get_running_builders(self):
        builders = []
        for bdict in (yield self.bot.getAllBuilders()):
            if (yield self.bot.getRunningBuilds(bdict['builderid'])):
                builders.append(bdict['name'])
        return builders

    @defer.inlineCallbacks
    @Contact.overrideCommand
    def command_WATCH(self, args, **kwargs):
        if args:
            super().command_WATCH(args)
        else:
            builders = yield self.get_running_builders()
            if builders:
                keyboard = [[self.query_button("🔎 " + b, f'/watch {b}')] for b in builders]
                self.send(
                    "Which builder do you want to watch?",
                    reply_markup={'inline_keyboard': keyboard},
                )
            else:
                self.send("There are no currently running builds.")

    @Contact.overrideCommand
    def command_NOTIFY(self, args, tquery=None, **kwargs):
        if args:
            want_list = args == 'list'
            if want_list and tquery:
                self.bot.delete_message(self.chat_id, tquery['message']['message_id'])

            super().command_NOTIFY(args)

            if want_list or not tquery:
                return

        keyboard = [
            [
                self.query_button(
                    f"{e.capitalize()} {'🔔' if e in self.channel.notify_events else '🔕'}",
                    f"/notify {'off' if e in self.channel.notify_events else 'on'}-quiet {e}",
                )
                for e in evs
            ]
            for evs in (
                ('started', 'finished'),
                ('success', 'failure'),
                ('warnings', 'exception'),
                ('problem', 'recovery'),
                ('worse', 'better'),
                ('cancelled', 'worker'),
            )
        ] + [[self.query_button("Hide...", '/notify list')]]

        if tquery:
            self.bot.edit_keyboard(self.chat_id, tquery['message']['message_id'], keyboard)
        else:
            self.send(
                "Here are available notifications and their current state. "
                "Click to turn them on/off.",
                reply_markup={'inline_keyboard': keyboard},
            )

    def ask_for_reply(self, prompt, greeting='Ok'):
        kwargs = {}
        if not self.is_private_chat:
            username = self.user_info.get('username', '')
            if username:
                if greeting:
                    prompt = f"{greeting} @{username}, now {prompt}..."
                else:
                    prompt = f"@{username}, now {prompt}..."
                kwargs['reply_markup'] = {'force_reply': True, 'selective': True}
            else:
                if greeting:
                    prompt = f"{greeting}, now reply to this message and {prompt}..."
                else:
                    prompt = f"Reply to this message and {prompt}..."
        else:
            if greeting:
                prompt = f"{greeting}, now {prompt}..."
            else:
                prompt = prompt[0].upper() + prompt[1:] + "..."
            # Telegram seems to have a bug, which causes reply request to pop sometimes again.
            # So we do not force reply to avoid it...
            # kwargs['reply_markup'] = {
            #     'force_reply': True
            # }
        self.send(prompt, **kwargs)

    @defer.inlineCallbacks
    @Contact.overrideCommand
    def command_STOP(self, args, **kwargs):
        argv = self.splitArgs(args)
        if len(argv) >= 3 or (argv and argv[0] != 'build'):
            super().command_STOP(args)
            return
        argv = argv[1:]
        if not argv:
            builders = yield self.get_running_builders()
            if builders:
                keyboard = [[self.query_button("🚫 " + b, f'/stop build {b}')] for b in builders]
                self.send("Select builder to stop...", reply_markup={'inline_keyboard': keyboard})
        else:  # len(argv) == 1
            self.template = '/stop ' + args + ' {}'
            self.ask_for_reply(f"give me the reason to stop build on `{argv[0]}`")

    @Contact.overrideCommand
    def command_SHUTDOWN(self, args, **kwargs):
        if args:
            return super().command_SHUTDOWN(args)
        if self.master.botmaster.shuttingDown:
            keyboard = [
                [
                    self.query_button("🔙 Stop Shutdown", '/shutdown stop'),
                    self.query_button("‼️ Shutdown Now", '/shutdown now'),
                ]
            ]
            text = "Buildbot is currently shutting down.\n\n"
        else:
            keyboard = [
                [
                    self.query_button("↘️ Begin Shutdown", '/shutdown start'),
                    self.query_button("‼️ Shutdown Now", '/shutdown now'),
                ]
            ]
            text = ""
        self.send(text + "What do you want to do?", reply_markup={'inline_keyboard': keyboard})
        return None

    @defer.inlineCallbacks
    def command_FORCE(self, args, tquery=None, partial=None, **kwargs):
        """force a build"""

        try:
            forceschedulers = yield self.master.data.get(('forceschedulers',))
        except AttributeError:
            forceschedulers = None
        else:
            forceschedulers = dict((s['name'], s) for s in forceschedulers)

        if not forceschedulers:
            raise UsageError("no force schedulers configured for use by /force")

        argv = self.splitArgs(args)

        try:
            sched = argv[0]
        except IndexError:
            if len(forceschedulers) == 1:
                sched = next(iter(forceschedulers))
            else:
                keyboard = [
                    [self.query_button(s['label'], f"/force {s['name']}")]
                    for s in forceschedulers.values()
                ]
                self.send(
                    "Which force scheduler do you want to activate?",
                    reply_markup={'inline_keyboard': keyboard},
                )
                return
        else:
            if sched in forceschedulers:
                del argv[0]
            elif len(forceschedulers) == 1:
                sched = next(iter(forceschedulers))
            else:
                raise UsageError(
                    f"Try '/force' and follow the instructions (no force scheduler {sched})"
                )
        scheduler = forceschedulers[sched]

        try:
            task = argv.pop(0)
        except IndexError:
            task = 'config'

        if tquery and task != 'config':
            self.bot.edit_keyboard(self.chat_id, tquery['message']['message_id'])

        if not argv:
            keyboard = [
                [self.query_button(b, f'/force {sched} {task} {b}')]
                for b in scheduler['builder_names']
            ]
            self.send(
                "Which builder do you want to start?", reply_markup={'inline_keyboard': keyboard}
            )
            return

        if task == 'ask':
            try:
                what = argv.pop(0)
            except IndexError as e:
                raise UsageError("Try '/force' and follow the instructions") from e
        else:
            what = None  # silence PyCharm warnings

        bldr = argv.pop(0)
        if bldr not in scheduler['builder_names']:
            raise UsageError(
                "Try '/force' and follow the instructions "
                f"(`{bldr}` not configured for _{scheduler['label']}_ scheduler)"
            )

        try:
            params = dict(arg.split('=', 1) for arg in argv)
        except ValueError as e:
            raise UsageError(f"Try '/force' and follow the instructions ({e})") from e

        all_fields = list(collect_fields(scheduler['all_fields']))
        required_params = [
            f['fullName']
            for f in all_fields
            if f['required'] and f['fullName'] not in ('username', 'owner')
        ]
        missing_params = [p for p in required_params if p not in params]

        if task == 'build':
            # TODO This should probably be moved to the upper class,
            # however, it will change the force command totally

            try:
                if missing_params:
                    # raise UsageError
                    task = 'config'
                else:
                    params.update(
                        dict(
                            (f['fullName'], f['default'])
                            for f in all_fields
                            if f['type'] == 'fixed' and f['fullName'] not in ('username', 'owner')
                        )
                    )

                    builder = yield self.bot.getBuilder(buildername=bldr)
                    for scheduler in self.master.allSchedulers():
                        if scheduler.name == sched and isinstance(scheduler, ForceScheduler):
                            break
                    else:
                        raise ValueError(f"There is no force scheduler '{sched}'")
                    try:
                        yield scheduler.force(
                            builderid=builder['builderid'], owner=self.describeUser(), **params
                        )
                    except CollectedValidationError as e:
                        raise ValueError(e.errors) from e
                    self.send("Force build successfully requested.")
                    return

            except (IndexError, ValueError) as e:
                raise UsageError(f"Try '/force' and follow the instructions ({e})") from e

        if task == 'config':
            msg = f"{self.user_full_name}, you are about to start a new build on `{bldr}`!"

            keyboard = []
            args = ' '.join(shlex.quote(f"{p[0]}={p[1]}") for p in params.items())

            fields = [
                f
                for f in all_fields
                if f['type'] != 'fixed' and f['fullName'] not in ('username', 'owner')
            ]

            if fields:
                msg += "\n\nThe current build parameters are:"
                for field in fields:
                    if field['type'] == 'nested':
                        msg += f"\n{field['label']}"
                    else:
                        field_name = field['fullName']
                        value = params.get(field_name, field['default'])
                        if isinstance(value, str):
                            value = value.strip()
                        msg += f"\n    {field['label']} `{value}`"
                        if value:
                            key = "Change "
                        else:
                            key = "Set "
                        key += field_name.replace('_', ' ').title()
                        if field_name in missing_params:
                            key = "⚠️ " + key
                            msg += " ⚠️"
                        keyboard.append([
                            self.query_button(key, f'/force {sched} ask {field_name} {bldr} {args}')
                        ])

            msg += "\n\nWhat do you want to do?"
            if missing_params:
                msg += " You must set values for all parameters marked with ⚠️"

            if not missing_params:
                keyboard.append(
                    [self.query_button("🚀 Start Build", f'/force {sched} build {bldr} {args}')],
                )

            self.send(msg, reply_markup={'inline_keyboard': keyboard})

        elif task == 'ask':
            prompt = "enter the new value for the " + what.replace('_', ' ').lower()
            args = ' '.join(shlex.quote(f"{p[0]}={p[1]}") for p in params.items() if p[0] != what)
            self.template = f'/force {sched} config {bldr} {args} {what}={{}}'
            self.ask_for_reply(prompt, '')

        else:
            raise UsageError("Try '/force' and follow the instructions")

    command_FORCE.usage = "force - Force a build"  # type: ignore[attr-defined]


class TelegramStatusBot(StatusBot):
    contactClass = TelegramContact
    channelClass = TelegramChannel
    commandPrefix = '/'

    offline_string = "offline ❌"
    idle_string = "idle 💤"
    running_string = "running 🌀:"

    query_cache: dict[int, dict[str, Any]] = {}

    @property
    def commandSuffix(self):
        if self.nickname is not None:
            return '@' + self.nickname
        return None

    def __init__(self, token, outgoing_http, chat_ids, *args, retry_delay=30, **kwargs):
        super().__init__(*args, **kwargs)

        self.http_client = outgoing_http
        self.retry_delay = retry_delay
        self.token = token

        self.chat_ids = chat_ids

        self.nickname = None

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        for c in self.chat_ids:
            channel = self.getChannel(c)
            channel.add_notification_events(self.notify_events)
        yield self.loadState()
        commands = [
            {'command': command, 'description': doc}
            for command, doc in TelegramContact.get_commands()
        ]
        self.post('/setMyCommands', json={'commands': commands})

    results_emoji = {
        SUCCESS: ' ✅',
        WARNINGS: ' ⚠️',
        FAILURE: '❗',
        EXCEPTION: ' ‼️',
        RETRY: ' 🔄',
        CANCELLED: ' 🚫',
    }

    def format_build_status(self, build, short=False):
        br = build['results']
        if short:
            return self.results_emoji[br]
        else:
            return self.results_descriptions[br] + self.results_emoji[br]

    def getContact(self, user, channel):
        """get a Contact instance for ``user`` on ``channel``"""
        assert isinstance(user, dict), "user must be a dict provided by Telegram API"
        assert isinstance(channel, dict), "channel must be a dict provided by Telegram API"

        uid = user['id']
        cid = channel['id']
        try:
            contact = self.contacts[(cid, uid)]
        except KeyError:
            valid = self.isValidUser(uid)
            contact = self.contactClass(user=user, channel=self.getChannel(channel, valid))
            if valid:
                self.contacts[(cid, uid)] = contact
        else:
            if isinstance(user, dict):
                contact.user_info.update(user)
            if isinstance(channel, dict):
                contact.channel.chat_info.update(channel)
        return contact

    def getChannel(self, channel, valid=True):
        if not isinstance(channel, dict):
            channel = {'id': channel}
        cid = channel['id']
        try:
            return self.channels[cid]
        except KeyError:
            new_channel = self.channelClass(self, channel)
            if valid:
                self.channels[cid] = new_channel
                new_channel.setServiceParent(self)
            return new_channel

    @defer.inlineCallbacks
    def process_update(self, update):
        data = {}

        message = update.get('message')
        if message is None:
            query = update.get('callback_query')
            if query is None:
                self.log('No message in Telegram update object')
                return 'no message'
            original_message = query.get('message', {})
            data = query.get('data', 0)
            try:
                data = self.query_cache[int(data)]
            except ValueError:
                text = data
                data = {}
                notify = None
            except KeyError:
                text = None
                data = {}
                notify = "Sorry, button is no longer valid!"
                if original_message:
                    try:
                        self.edit_keyboard(
                            original_message['chat']['id'], original_message['message_id']
                        )
                    except KeyError:
                        pass
            else:
                if isinstance(data, dict):
                    data = data.copy()
                    text = data.pop('command')
                    try:
                        notify = data.pop('notify')
                    except KeyError:
                        notify = None
                else:
                    text = data
                    data = {}
                    notify = None
            data['tquery'] = query
            self.answer_query(query['id'], notify)
            message = {
                'from': query['from'],
                'chat': original_message.get('chat'),
                'text': text,
            }
            if 'reply_to_message' in original_message:
                message['reply_to_message'] = original_message['reply_to_message']

        chat = message['chat']

        user = message.get('from')
        if user is None:
            self.log('No user in incoming message')
            return 'no user'

        text = message.get('text')
        if not text:
            return 'no text in the message'

        contact = self.getContact(user=user, channel=chat)
        data['tmessage'] = message
        template = contact.template
        contact.template = None
        if text.startswith(self.commandPrefix):
            result = yield contact.handleMessage(text, **data)
        else:
            if template:
                text = template.format(shlex.quote(text))
            result = yield contact.handleMessage(text, **data)
        return result

    @defer.inlineCallbacks
    def post(self, path, **kwargs):
        logme = True
        while True:
            try:
                res = yield self.http_client.post(path, **kwargs)
            except AssertionError as err:
                # just for tests
                raise err
            except Exception as err:
                msg = f"ERROR: problem sending Telegram request {path} (will try again): {err}"
                if logme:
                    self.log(msg)
                    logme = False
                yield asyncSleep(self.retry_delay)
            else:
                ans = yield res.json()
                if not ans.get('ok'):
                    self.log(
                        f"ERROR: cannot send Telegram request {path}: "
                        f"[{res.code}] {ans.get('description')}"
                    )
                    return None
                return ans.get('result', True)

    @defer.inlineCallbacks
    def set_nickname(self):
        res = yield self.post('/getMe')
        if res:
            self.nickname = res.get('username')

    @defer.inlineCallbacks
    def answer_query(self, query_id, notify=None):
        params = {"callback_query_id": query_id}
        if notify is not None:
            params.update({"text": notify})
        return (yield self.post('/answerCallbackQuery', json=params))

    @defer.inlineCallbacks
    def send_message(
        self,
        chat,
        message,
        parse_mode='Markdown',
        reply_to_message_id=None,
        reply_markup=None,
        **kwargs,
    ):
        result = None

        message = message.strip()
        while message:
            params = {"chat_id": chat}
            if parse_mode is not None:
                params['parse_mode'] = parse_mode
            if reply_to_message_id is not None:
                params['reply_to_message_id'] = reply_to_message_id
                reply_to_message_id = None  # we only mark first message as a reply

            if len(message) <= 4096:
                params['text'] = message
                message = None
            else:
                n = message[:4096].rfind('\n')
                n = n + 1 if n != -1 else 4096
                params['text'] = message[:n].rstrip()
                message = message[n:].lstrip()

            if not message and reply_markup is not None:
                params['reply_markup'] = reply_markup

            params.update(kwargs)

            result = yield self.post('/sendMessage', json=params)

        return result

    @defer.inlineCallbacks
    def edit_message(self, chat, msg, message, parse_mode='Markdown', **kwargs):
        params = {"chat_id": chat, "message_id": msg, "text": message}
        if parse_mode is not None:
            params['parse_mode'] = parse_mode
        params.update(kwargs)
        return (yield self.post('/editMessageText', json=params))

    @defer.inlineCallbacks
    def edit_keyboard(self, chat, msg, keyboard=None):
        params = {"chat_id": chat, "message_id": msg}
        if keyboard is not None:
            params['reply_markup'] = {'inline_keyboard': keyboard}
        return (yield self.post('/editMessageReplyMarkup', json=params))

    @defer.inlineCallbacks
    def delete_message(self, chat, msg):
        params = {"chat_id": chat, "message_id": msg}
        return (yield self.post('/deleteMessage', json=params))

    @defer.inlineCallbacks
    def send_sticker(self, chat, sticker, **kwargs):
        params = {"chat_id": chat, "sticker": sticker}
        params.update(kwargs)
        return (yield self.post('/sendSticker', json=params))


class TelegramWebhookBot(TelegramStatusBot):
    name: str | None = "TelegramWebhookBot"  # type: ignore[assignment]

    def __init__(self, token, *args, certificate=None, **kwargs):
        TelegramStatusBot.__init__(self, token, *args, **kwargs)
        self._certificate = certificate
        self.webhook = WebhookResource('telegram' + token)
        self.webhook.setServiceParent(self)

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        url = bytes2unicode(self.master.config.buildbotURL)
        if not url.endswith('/'):
            url += '/'
        yield self.set_webhook(url + self.webhook.path, self._certificate)

    def process_webhook(self, request):
        update = self.get_update(request)
        return self.process_update(update)

    def get_update(self, request):
        content = request.content.read()
        content = bytes2unicode(content)
        content_type = request.getHeader(b'Content-Type')
        content_type = bytes2unicode(content_type)
        if content_type is not None and content_type.startswith('application/json'):
            update = json.loads(content)
        else:
            raise ValueError(f'Unknown content type: {content_type}')
        return update

    @defer.inlineCallbacks
    def set_webhook(self, url, certificate=None):
        if not certificate:
            self.log(f"Setting up webhook to: {url}")
            yield self.post('/setWebhook', json={"url": url})
        else:
            self.log(f"Setting up webhook to: {url} (custom certificate)")
            certificate = io.BytesIO(unicode2bytes(certificate))
            yield self.post('/setWebhook', data={"url": url}, files={"certificate": certificate})


class TelegramPollingBot(TelegramStatusBot):
    name: str | None = "TelegramPollingBot"  # type: ignore[assignment]

    def __init__(self, *args, poll_timeout=120, **kwargs):
        super().__init__(*args, **kwargs)
        self._polling_finished_notifier = Notifier()
        self.poll_timeout = poll_timeout

    def startService(self):
        super().startService()
        self._polling_continue = True
        self.do_polling()

    @defer.inlineCallbacks
    def stopService(self):
        self._polling_continue = False
        yield self._polling_finished_notifier.wait()
        yield super().stopService()

    @defer.inlineCallbacks
    def do_polling(self):
        yield self.post('/deleteWebhook')
        offset = 0
        kwargs = {'json': {'timeout': self.poll_timeout}}
        logme = True
        while self._polling_continue:
            if offset:
                kwargs['json']['offset'] = offset
            try:
                res = yield self.http_client.post(
                    '/getUpdates', timeout=self.poll_timeout + 2, **kwargs
                )
                ans = yield res.json()
                if not ans.get('ok'):
                    raise ValueError(f"[{res.code}] {ans.get('description')}")
                updates = ans.get('result')
            except AssertionError as err:
                raise err
            except Exception as err:
                msg = f"ERROR: cannot send Telegram request /getUpdates (will try again): {err}"
                if logme:
                    self.log(msg)
                    logme = False
                yield asyncSleep(self.retry_delay)
            else:
                logme = True
                if updates:
                    offset = max(update['update_id'] for update in updates) + 1
                    for update in updates:
                        yield self.process_update(update)

        self._polling_finished_notifier.notify(None)


class TelegramBot(service.BuildbotService):
    name = "TelegramBot"

    in_test_harness = False

    compare_attrs: ClassVar[Sequence[str]] = [
        "bot_token",
        "chat_ids",
        "authz",
        "tags",
        "notify_events",
        "showBlameList",
        "useRevisions",
        "certificate",
        "useWebhook",
        "pollTimeout",
        "retryDelay",
    ]
    secrets = ["bot_token"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bot = None

    def _get_http(self, bot_token):
        base_url = "https://api.telegram.org/bot" + bot_token
        return httpclientservice.HTTPSession(self.master.httpservice, base_url)

    def checkConfig(
        self,
        bot_token,
        chat_ids=None,
        authz=None,
        bot_username=None,
        tags=None,
        notify_events=None,
        showBlameList=True,
        useRevisions=False,
        useWebhook=False,
        certificate=None,
        pollTimeout=120,
        retryDelay=30,
    ):
        super().checkConfig(self.name)

        if authz is not None:
            for acl in authz.values():
                if not isinstance(acl, (list, tuple, bool)):
                    config.error("authz values must be bool or a list of user ids")

        if isinstance(certificate, io.TextIOBase):
            config.error("certificate file must be open in binary mode")

    @defer.inlineCallbacks
    def reconfigService(
        self,
        bot_token,
        chat_ids=None,
        authz=None,
        bot_username=None,
        tags=None,
        notify_events=None,
        showBlameList=True,
        useRevisions=False,
        useWebhook=False,
        certificate=None,
        pollTimeout=120,
        retryDelay=30,
    ):
        # need to stash these so we can detect changes later
        self.bot_token = bot_token
        if chat_ids is None:
            chat_ids = []
        self.chat_ids = chat_ids
        self.authz = authz
        self.useRevisions = useRevisions
        self.tags = tags
        if notify_events is None:
            notify_events = set()
        self.notify_events = notify_events
        self.useWebhook = useWebhook
        self.certificate = certificate
        self.pollTimeout = pollTimeout
        self.retryDelay = retryDelay

        # This function is only called in case of reconfig with changes
        # We don't try to be smart here. Just restart the bot if config has
        # changed.

        http = yield self._get_http(bot_token)

        if self.bot is not None:
            self.removeService(self.bot)

        if not useWebhook:
            self.bot = TelegramPollingBot(
                bot_token,
                http,
                chat_ids,
                authz,
                tags=tags,
                notify_events=notify_events,
                useRevisions=useRevisions,
                showBlameList=showBlameList,
                poll_timeout=self.pollTimeout,
                retry_delay=self.retryDelay,
            )
        else:
            self.bot = TelegramWebhookBot(
                bot_token,
                http,
                chat_ids,
                authz,
                tags=tags,
                notify_events=notify_events,
                useRevisions=useRevisions,
                showBlameList=showBlameList,
                retry_delay=self.retryDelay,
                certificate=self.certificate,
            )
        if bot_username is not None:
            self.bot.nickname = bot_username
        else:
            yield self.bot.set_nickname()
            if self.bot.nickname is None:
                raise RuntimeError("No bot username specified and I cannot get it from Telegram")

        yield self.bot.setServiceParent(self)



# Copyright Buildbot Team Members
# Portions copyright 2015-2016 ClusterHQ Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.base import _ThreePhaseEvent
from twisted.internet.interfaces import IReactorCore
from twisted.internet.interfaces import IReactorThreads
from twisted.internet.task import Clock
from twisted.python import log
from twisted.python.failure import Failure
from zope.interface import implementer

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence

if TYPE_CHECKING:
    from typing import Any
    from typing import TypeVar

    from twisted.internet.interfaces import IDelayedCall
    from twisted.internet.interfaces import IReactorTime
    from typing_extensions import ParamSpec

    from buildbot.util.twisted import ThreadPool
    from buildbot_worker.util.twisted import InlineCallbacksType

    _T = TypeVar('_T')
    _P = ParamSpec('_P')


# The code here is based on the implementations in
# https://twistedmatrix.com/trac/ticket/8295
# https://twistedmatrix.com/trac/ticket/8296


@implementer(IReactorCore)
class CoreReactor:
    """
    Partial implementation of ``IReactorCore``.
    """

    def __init__(self) -> None:
        super().__init__()
        self._triggers: dict[str, _ThreePhaseEvent] = {}

    def addSystemEventTrigger(
        self, phase: str, eventType: str, callable: Callable, *args: object, **kw: object
    ) -> Any:
        event = self._triggers.setdefault(eventType, _ThreePhaseEvent())
        return eventType, event.addTrigger(phase, callable, *args, **kw)

    def removeSystemEventTrigger(self, triggerID: tuple[str, Any]) -> None:
        eventType, handle = triggerID
        event = self._triggers.setdefault(eventType, _ThreePhaseEvent())
        event.removeTrigger(handle)

    def fireSystemEvent(self, eventType: str) -> None:
        event = self._triggers.get(eventType)
        if event is not None:
            event.fireEvent()

    def callWhenRunning(self, callable: Callable, *args: object, **kwargs: object) -> Any | None:
        callable(*args, **kwargs)
        return None

    def crash(self) -> None:
        raise NotImplementedError()

    def iterate(self, delay: float = 0) -> None:
        raise NotImplementedError()

    def run(self) -> None:
        raise NotImplementedError()

    def running(self) -> bool:
        raise NotImplementedError()

    def resolve(self, name: str, timeout: Sequence[int]) -> defer.Deferred[str]:  # type: ignore[override]
        raise NotImplementedError()

    def stop(self) -> None:
        raise NotImplementedError()


class NonThreadPool:
    """
    A stand-in for ``twisted.python.threadpool.ThreadPool`` so that the
    majority of the test suite does not need to use multithreading.

    This implementation takes the function call which is meant to run in a
    thread pool and runs it synchronously in the calling thread.

    :ivar int calls: The number of calls which have been dispatched to this
        object.
    """

    calls = 0

    def __init__(self, **kwargs: Any) -> None:
        pass

    def callInThreadWithCallback(
        self,
        onResult: Callable[[bool, Failure | _T], None],
        func: Callable[_P, _T],
        *args: _P.args,
        **kw: _P.kwargs,
    ) -> None:
        self.calls += 1
        try:
            result = func(*args, **kw)
        except:  # noqa: E722
            # We catch *everything* here, since normally this code would be
            # running in a thread, where there is nothing that will catch
            # error.
            onResult(False, Failure())
        else:
            onResult(True, result)

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass


@implementer(IReactorThreads)
class NonReactor:
    """
    A partial implementation of ``IReactorThreads`` which fits into
    the execution model defined by ``NonThreadPool``.
    """

    def suggestThreadPoolSize(self, size: int) -> None:
        # we don't do threads, so this is a no-op
        pass

    def callFromThread(self, callable: Callable, *args: object, **kwargs: object) -> None:
        callable(*args, **kwargs)
        return None

    def callInThread(self, callable: Callable, *args: object, **kwargs: object) -> None:
        callable(*args, **kwargs)
        return None

    def getThreadPool(self) -> ThreadPool:
        return cast("ThreadPool", NonThreadPool())


class TestReactor(NonReactor, CoreReactor, Clock):
    def __init__(self) -> None:
        super().__init__()

        # whether there are calls that should run right now
        self._pendingCurrentCalls = False
        self.stop_called = False

    def _executeCurrentDelayedCalls(self) -> None:
        while self.getDelayedCalls():
            first = sorted(self.getDelayedCalls(), key=lambda a: a.getTime())[0]
            if first.getTime() > self.seconds():
                break
            self.advance(0)

        self._pendingCurrentCalls = False

    @defer.inlineCallbacks
    def _catchPrintExceptions(
        self,
        what: Callable[_P, None | defer.Deferred[None]],
        *a: _P.args,
        **kw: _P.kwargs,
    ) -> InlineCallbacksType[None]:
        try:
            r = what(*a, **kw)
            if isinstance(r, defer.Deferred):
                yield r
        except Exception as e:
            log.msg('Unhandled exception from deferred when doing TestReactor.advance()', e)
            raise

    def callLater(
        self,
        delay: float,
        callable: Callable[..., object],
        *args: object,
        **kw: object,
    ) -> IDelayedCall:
        # Buildbot often uses callLater(0, ...) to defer execution of certain
        # code to the next iteration of the reactor. This means that often
        # there are pending callbacks registered to the reactor that might
        # block other code from proceeding unless the test reactor has an
        # iteration. To avoid deadlocks in tests we give the real reactor a
        # chance to advance the test reactor whenever we detect that there
        # are callbacks that should run in the next iteration of the test
        # reactor.
        #
        # Additionally, we wrap all calls with a function that prints any
        # unhandled exceptions
        if delay <= 0 and not self._pendingCurrentCalls:
            cast("IReactorTime", reactor).callLater(0, self._executeCurrentDelayedCalls)

        return super().callLater(delay, self._catchPrintExceptions, callable, *args, **kw)

    def stop(self) -> None:
        # first fire pending calls until the current time. Note that the real
        # reactor only advances until the current time in the case of shutdown.
        self.advance(0)

        # then, fire the shutdown event
        self.fireSystemEvent('shutdown')

        self.stop_called = True



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.changes import base
from buildbot.pbutil import NewCredPerspective

if TYPE_CHECKING:
    from buildbot.config.master import MasterConfig
    from buildbot.util.twisted import InlineCallbacksType


class ChangePerspective(NewCredPerspective):
    def __init__(self, master: Any, prefix: str | None):
        self.master = master
        self.prefix = prefix

    def attached(self, mind: Any) -> Any:
        return self

    def detached(self, mind: Any) -> None:
        pass

    @defer.inlineCallbacks
    def perspective_addChange(self, changedict: dict[str, Any]) -> InlineCallbacksType[None]:
        log.msg("perspective_addChange called")

        if 'revlink' in changedict and not changedict['revlink']:
            changedict['revlink'] = ''
        if 'repository' in changedict and not changedict['repository']:
            changedict['repository'] = ''
        if 'project' in changedict and not changedict['project']:
            changedict['project'] = ''
        if 'files' not in changedict or not changedict['files']:
            changedict['files'] = []
        if 'committer' in changedict and not changedict['committer']:
            changedict['committer'] = None

        # rename arguments to new names.  Note that the client still uses the
        # "old" names (who, when, and isdir), as they are not deprecated yet,
        # although the master will accept the new names (author,
        # when_timestamp).  After a few revisions have passed, we
        # can switch the client to use the new names.
        if 'who' in changedict:
            changedict['author'] = changedict['who']
            del changedict['who']
        if 'when' in changedict:
            changedict['when_timestamp'] = changedict['when']
            del changedict['when']

        # turn any bytestring keys into unicode, assuming utf8 but just
        # replacing unknown characters.  Ideally client would send us unicode
        # in the first place, but older clients do not, so this fallback is
        # useful.
        for key, value in changedict.items():
            if isinstance(value, bytes):
                changedict[key] = value.decode('utf8', 'replace')
        changedict['files'] = list(changedict['files'])
        for i, file in enumerate(changedict.get('files', [])):
            if isinstance(file, bytes):
                changedict['files'][i] = file.decode('utf8', 'replace')

        files = []
        for path in changedict['files']:
            if self.prefix:
                if not path.startswith(self.prefix):
                    # this file does not start with the prefix, so ignore it
                    continue
                path = path[len(self.prefix) :]
            files.append(path)
        changedict['files'] = files

        if not files:
            log.msg("No files listed in change... bit strange, but not fatal.")

        if "links" in changedict:
            log.msg("Found links: " + repr(changedict['links']))
            del changedict['links']

        yield self.master.data.updates.addChange(**changedict)


class PBChangeSource(base.ChangeSource):
    compare_attrs: ClassVar[Sequence[str]] = ("user", "passwd", "port", "prefix", "port")

    def __init__(
        self,
        user: str = "change",
        passwd: str = "changepw",
        port: int | str | None = None,
        prefix: str | None = None,
        name: str | None = None,
    ):
        if name is None:
            if prefix:
                name = f"PBChangeSource:{prefix}:{port}"
            else:
                name = f"PBChangeSource:{port}"

        super().__init__(name=name)

        self.user = user
        self.passwd = passwd
        self.port = port
        self.prefix = prefix
        self.registration: Any = None
        self.registered_port: int | str | None = None

    def describe(self) -> str:
        portname = self.registered_port
        d = "PBChangeSource listener on " + str(portname)
        if self.prefix is not None:
            d += f" (prefix '{self.prefix}')"
        return d

    def _calculatePort(self, cfg: MasterConfig) -> int | str | None:
        # calculate the new port, defaulting to the worker's PB port if
        # none was specified
        port = self.port
        if port is None:
            port = cfg.protocols.get('pb', {}).get('port')
        return port

    @defer.inlineCallbacks
    def reconfigServiceWithBuildbotConfig(
        self, new_config: MasterConfig
    ) -> InlineCallbacksType[None]:  # type: ignore[override]
        port = self._calculatePort(new_config)
        if not port:
            config.error("No port specified for PBChangeSource, and no worker port configured")

        # and, if it's changed, re-register
        if port != self.registered_port and self.isActive():
            yield self._unregister()
            yield self._register(port)

        yield super().reconfigServiceWithBuildbotConfig(new_config)

    @defer.inlineCallbacks
    def activate(self) -> InlineCallbacksType[None]:
        port = self._calculatePort(self.master.config)
        yield self._register(port)

    def deactivate(self) -> defer.Deferred[None]:
        return self._unregister()

    @defer.inlineCallbacks
    def _register(self, port: int | str | None) -> InlineCallbacksType[None]:
        if not port:
            return
        self.registered_port = port
        self.registration = yield self.master.pbmanager.register(
            port, self.user, self.passwd, self.getPerspective
        )

    def _unregister(self) -> defer.Deferred[None]:
        self.registered_port = None
        if self.registration:
            reg = self.registration
            self.registration = None
            return reg.unregister()
        return defer.succeed(None)

    def getPerspective(self, mind: Any, username: str) -> ChangePerspective:
        assert username == self.user
        return ChangePerspective(self.master, self.prefix)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This module is left for backward compatibility of old-named worker API.
# It should never be imported by Buildbot.
from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.python import runtime

if TYPE_CHECKING:
    from typing import Callable
    from typing import Literal
    from typing import TypeVar

    _T = TypeVar('_T')


def skipUnlessPlatformIs(platform: Literal["posix", "win32"]) -> Callable[[_T], _T]:
    def closure(test: _T) -> _T:
        if runtime.platformType != platform:
            test.skip = f"not a {platform} platform"  # type: ignore[attr-defined]
        return test

    return closure



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import io
import os
import re
import shutil
import tarfile
from typing import TYPE_CHECKING
from typing import cast

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import failure
from twisted.python import runtime
from twisted.trial import unittest

from buildbot_worker.commands import transfer
from buildbot_worker.test.fake.remote import FakeRemote
from buildbot_worker.test.util.command import CommandTestMixin

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable

    from twisted.internet.interfaces import IReactorTime

    from buildbot_worker.util.twisted import InlineCallbacksType


class FakeMasterMethods:
    # a fake to represent any of:
    # - FileWriter
    # - FileDirectoryWriter
    # - FileReader

    def __init__(self, add_update: Callable[[tuple[str, Any] | str], None]) -> None:
        self.add_update = add_update

        self.delay_write = False
        self.count_writes = False
        self.keep_data = False
        self.write_out_of_space_at: int | None = None

        self.delay_read = False
        self.count_reads = False

        self.unpack_fail = False

        self.written = False
        self.read = False
        self.data = b''

    def remote_write(self, data: bytes) -> defer.Deferred[None] | None:
        if self.write_out_of_space_at is not None:
            self.write_out_of_space_at -= len(data)
            if self.write_out_of_space_at <= 0:
                f = failure.Failure(RuntimeError("out of space"))
                return defer.fail(f)
        if self.count_writes:
            self.add_update(f'write {len(data)}')
        elif not self.written:
            self.add_update('write(s)')
            self.written = True

        if self.keep_data:
            self.data += data

        if self.delay_write:
            d: defer.Deferred[None] = defer.Deferred()
            cast("IReactorTime", reactor).callLater(0.01, d.callback, None)
            return d
        return None

    def remote_read(self, length: int) -> defer.Deferred[bytes] | bytes | str:
        if self.count_reads:
            self.add_update(f'read {length}')
        elif not self.read:
            self.add_update('read(s)')
            self.read = True

        if not self.data:
            return ''

        _slice, self.data = self.data[:length], self.data[length:]
        if self.delay_read:
            d: defer.Deferred[bytes] = defer.Deferred()
            cast("IReactorTime", reactor).callLater(0.01, d.callback, _slice)
            return d
        return _slice

    def remote_unpack(self) -> defer.Deferred[None] | None:
        self.add_update('unpack')
        if self.unpack_fail:
            return defer.fail(failure.Failure(RuntimeError("out of space")))
        return None

    def remote_utime(self, accessed_modified: tuple[float]) -> None:
        self.add_update(f'utime - {accessed_modified[0]}')

    def remote_close(self) -> None:
        self.add_update('close')


class TestUploadFile(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # write 180 bytes of data to upload
        self.datadir = os.path.join(self.basedir, 'workdir')
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)
        os.makedirs(self.datadir)

        self.datafile = os.path.join(self.datadir, 'data')
        # note: use of 'wb' here ensures newlines aren't translated on the
        # upload
        with open(self.datafile, mode="wb") as f:
            f.write(b"this is some data\n" * 10)

    def tearDown(self) -> None:
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        self.fakemaster.count_writes = True  # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 1000,
                'blocksize': 64,
                'keepstamp': False,
            },
        )

        yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datafile}\n'),
            'write 64',
            'write 64',
            'write 52',
            'close',
            ('rc', 0),
        ])

    @defer.inlineCallbacks
    def test_truncated(self) -> InlineCallbacksType[None]:
        self.fakemaster.count_writes = True  # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 100,
                'blocksize': 64,
                'keepstamp': False,
            },
        )

        yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datafile}\n'),
            'write 64',
            'write 36',
            'close',
            ('rc', 1),
            ('stderr', f"Maximum filesize reached, truncating file '{self.datafile}'"),
        ])

    @defer.inlineCallbacks
    def test_missing(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data-nosuch'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 100,
                'blocksize': 64,
                'keepstamp': False,
            },
        )

        yield self.run_command()

        df = self.datafile + "-nosuch"
        self.assertUpdates([
            ('header', f'sending {df}\n'),
            'close',
            ('rc', 1),
            ('stderr', f"Cannot open file '{df}' for upload"),
        ])

    @defer.inlineCallbacks
    def test_out_of_space(self) -> InlineCallbacksType[None]:
        self.fakemaster.write_out_of_space_at = 70
        self.fakemaster.count_writes = True  # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 1000,
                'blocksize': 64,
                'keepstamp': False,
            },
        )

        with self.assertRaises(RuntimeError):
            yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datafile}\n'),
            'write 64',
            'close',
            ('rc', 1),
        ])

    @defer.inlineCallbacks
    def test_interrupted(self) -> InlineCallbacksType[None]:
        self.fakemaster.delay_write = True  # write very slowly

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 100,
                'blocksize': 2,
                'keepstamp': False,
            },
        )

        d = self.run_command()

        # wait a jiffy..
        interrupt_d: defer.Deferred[None] = defer.Deferred()
        cast("IReactorTime", reactor).callLater(0.01, interrupt_d.callback, None)

        # and then interrupt the step
        def do_interrupt(_: Any) -> defer.Deferred[None] | None:
            return self.cmd.interrupt()

        interrupt_d.addCallback(do_interrupt)

        yield defer.DeferredList([d, interrupt_d], consumeErrors=True)

        self.assertUpdates([
            ('header', f'sending {self.datafile}\n'),
            'write(s)',
            'close',
            ('rc', 1),
        ])

    @defer.inlineCallbacks
    def test_timestamp(self) -> InlineCallbacksType[None]:
        self.fakemaster.count_writes = True  # get actual byte counts
        timestamp = (os.path.getatime(self.datafile), os.path.getmtime(self.datafile))

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileUploadCommand,
            {
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': 1000,
                'blocksize': 64,
                'keepstamp': True,
            },
        )

        yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datafile}\n'),
            'write 64',
            'write 64',
            'write 52',
            'close',
            f'utime - {timestamp[0]}',
            ('rc', 0),
        ])


class TestWorkerDirectoryUpload(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # write a directory to upload
        self.datadir = os.path.join(self.basedir, 'workdir', 'data')
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)
        os.makedirs(self.datadir)
        with open(os.path.join(self.datadir, "aa"), mode="wb") as f:
            f.write(b"lots of a" * 100)
        with open(os.path.join(self.datadir, "bb"), mode="wb") as f:
            f.write(b"and a little b" * 17)

    def tearDown(self) -> None:
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)

    @defer.inlineCallbacks
    def test_simple(self, compress: str | None = None) -> InlineCallbacksType[None]:
        self.fakemaster.keep_data = True
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerDirectoryUploadCommand,
            {
                'workdir': 'workdir',
                'path': path,
                'writer': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 512,
                'compress': compress,
            },
        )

        yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datadir}\n'),
            'write(s)',
            'unpack',  # note no 'close"
            ('rc', 0),
        ])

        f = io.BytesIO(self.fakemaster.data)
        a = tarfile.open(fileobj=f, name='check.tar', mode="r")
        exp_names = ['.', 'aa', 'bb']
        got_names = [n.rstrip('/') for n in a.getnames()]
        # py27 uses '' instead of '.'
        got_names = sorted([n or '.' for n in got_names])
        self.assertEqual(got_names, exp_names, "expected archive contents")
        a.close()
        f.close()

    # try it again with bz2 and gzip
    def test_simple_bz2(self) -> defer.Deferred[None]:
        return self.test_simple('bz2')

    def test_simple_gz(self) -> defer.Deferred[None]:
        return self.test_simple('gz')

    # except bz2 can't operate in stream mode on py24

    @defer.inlineCallbacks
    def test_out_of_space_unpack(self) -> InlineCallbacksType[None]:
        self.fakemaster.keep_data = True
        self.fakemaster.unpack_fail = True
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))

        self.make_command(
            transfer.WorkerDirectoryUploadCommand,
            {
                'path': path,
                'workersrc': 'data',
                'writer': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 512,
                'compress': None,
            },
        )

        with self.assertRaises(RuntimeError):
            yield self.run_command()

        self.assertUpdates([
            ('header', f'sending {self.datadir}\n'),
            'write(s)',
            'unpack',
            ('rc', 1),
        ])


class TestWorkerDirectoryUploadNoDir(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()
        self.fakemaster = FakeMasterMethods(self.add_update)

    @defer.inlineCallbacks
    def test_directory_not_available(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))

        self.make_command(
            transfer.WorkerDirectoryUploadCommand,
            {
                'path': path,
                'workersrc': 'data',
                'writer': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 512,
                'compress': None,
            },
        )

        yield self.run_command()

        updates = self.get_updates()
        self.assertEqual(updates[0], ('rc', 1))
        self.assertEqual(updates[1][0], "stderr")

        error_msg = updates[1][1]
        pattern = re.compile("Cannot read directory (.*?) for upload: (.*?)")
        match = pattern.match(error_msg)
        self.assertNotEqual(error_msg, match)


class TestDownloadFile(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # the command will write to the basedir, so make sure it exists
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)
        os.makedirs(self.basedir)

    def tearDown(self) -> None:
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        self.fakemaster.count_reads = True  # get actual byte counts
        self.fakemaster.data = test_data = b'1234' * 13
        assert len(self.fakemaster.data) == 52

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileDownloadCommand,
            {
                'path': path,
                'reader': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 32,
                'mode': 0o777,
            },
        )

        yield self.run_command()

        self.assertUpdates(['read 32', 'read 32', 'read 32', 'close', ('rc', 0)])
        datafile = os.path.join(self.basedir, 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            datafileContent = f.read()
        self.assertEqual(datafileContent, test_data)
        if runtime.platformType != 'win32':
            self.assertEqual(os.stat(datafile).st_mode & 0o777, 0o777)

    @defer.inlineCallbacks
    def test_mkdir(self) -> InlineCallbacksType[None]:
        self.fakemaster.data = test_data = b'hi'

        path = os.path.join(
            self.basedir, 'workdir', os.path.expanduser(os.path.join('subdir', 'data'))
        )
        self.make_command(
            transfer.WorkerFileDownloadCommand,
            {
                'path': path,
                'reader': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 32,
                'mode': 0o777,
            },
        )

        yield self.run_command()

        self.assertUpdates(['read(s)', 'close', ('rc', 0)])
        datafile = os.path.join(self.basedir, 'workdir', 'subdir', 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            datafileContent = f.read()
        self.assertEqual(datafileContent, test_data)

    @defer.inlineCallbacks
    def test_failure(self) -> InlineCallbacksType[None]:
        self.fakemaster.data = b'hi'

        os.makedirs(os.path.join(self.basedir, 'dir'))
        path = os.path.join(self.basedir, os.path.expanduser('dir'))
        self.make_command(
            transfer.WorkerFileDownloadCommand,
            {
                'path': path,
                'reader': FakeRemote(self.fakemaster),
                'maxsize': None,
                'blocksize': 32,
                'mode': 0o777,
            },
        )

        yield self.run_command()

        self.assertUpdates([
            'close',
            ('rc', 1),
            (
                'stderr',
                "Cannot open file '{}' for download".format(os.path.join(self.basedir, 'dir')),
            ),
        ])

    @defer.inlineCallbacks
    def test_truncated(self) -> InlineCallbacksType[None]:
        self.fakemaster.data = test_data = b'tenchars--' * 10

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileDownloadCommand,
            {
                'path': path,
                'reader': FakeRemote(self.fakemaster),
                'maxsize': 50,
                'blocksize': 32,
                'mode': 0o777,
            },
        )
        yield self.run_command()

        self.assertUpdates([
            'read(s)',
            'close',
            ('rc', 1),
            (
                'stderr',
                "Maximum filesize reached, truncating file '{}'".format(
                    os.path.join(self.basedir, 'data')
                ),
            ),
        ])
        datafile = os.path.join(self.basedir, 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            data = f.read()
        self.assertEqual(data, test_data[:50])

    @defer.inlineCallbacks
    def test_interrupted(self) -> InlineCallbacksType[None]:
        self.fakemaster.data = b'tenchars--' * 100  # 1k
        self.fakemaster.delay_read = True  # read very slowly

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(
            transfer.WorkerFileDownloadCommand,
            {
                'path': path,
                'reader': FakeRemote(self.fakemaster),
                'maxsize': 100,
                'blocksize': 2,
                'mode': 0o777,
            },
        )

        d = self.run_command()

        # wait a jiffy..
        interrupt_d: defer.Deferred[None] = defer.Deferred()
        cast("IReactorTime", reactor).callLater(0.01, interrupt_d.callback, None)

        # and then interrupt the step
        def do_interrupt(_: Any) -> defer.Deferred[None] | None:
            return self.cmd.interrupt()

        interrupt_d.addCallback(do_interrupt)

        yield defer.DeferredList([d, interrupt_d], consumeErrors=True)

        self.assertUpdates(['read(s)', 'close', ('rc', 1)])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized
from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.spread import pb

from buildbot.config import BuilderConfig
from buildbot.config.master import MasterConfig
from buildbot.interfaces import LatentWorkerCannotSubstantiate
from buildbot.interfaces import LatentWorkerFailedToSubstantiate
from buildbot.interfaces import LatentWorkerSubstantiatiationCancelled
from buildbot.machine.latent import States as MachineStates
from buildbot.process.factory import BuildFactory
from buildbot.process.properties import Interpolate
from buildbot.process.properties import Properties
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.test.fake.latent import ControllableLatentWorker
from buildbot.test.fake.latent import LatentController
from buildbot.test.fake.machine import LatentMachineController
from buildbot.test.fake.step import BuildStepController
from buildbot.test.util.integration import RunFakeMasterTestCase
from buildbot.test.util.misc import TimeoutableTestCase
from buildbot.test.util.patch_delay import patchForDelay
from buildbot.worker import manager
from buildbot.worker.latent import States


class TestException(Exception):
    """
    An exception thrown in tests.
    """


class Latent(TimeoutableTestCase, RunFakeMasterTestCase):
    def tearDown(self):
        # Flush the errors logged by the master stop cancelling the builds.
        self.flushLoggedErrors(LatentWorkerSubstantiatiationCancelled)
        super().tearDown()

    @defer.inlineCallbacks
    def create_single_worker_config(self, controller_kwargs=None):
        if not controller_kwargs:
            controller_kwargs = {}

        controller = LatentController(self, 'local', **controller_kwargs)
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy",
                    workernames=["local"],
                    factory=BuildFactory(),
                ),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_id = yield self.master.data.updates.findBuilderId('testy')

        return controller, builder_id

    @defer.inlineCallbacks
    def create_single_worker_config_with_step(self, controller_kwargs=None):
        if not controller_kwargs:
            controller_kwargs = {}

        controller = LatentController(self, 'local', **controller_kwargs)
        stepcontroller = BuildStepController()
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy",
                    workernames=["local"],
                    factory=BuildFactory([stepcontroller.step]),
                ),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_id = yield self.master.data.updates.findBuilderId('testy')

        return controller, stepcontroller, builder_id

    @defer.inlineCallbacks
    def create_single_worker_two_builder_config(self, controller_kwargs=None):
        if not controller_kwargs:
            controller_kwargs = {}

        controller = LatentController(self, 'local', **controller_kwargs)
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy-1",
                    workernames=["local"],
                    factory=BuildFactory(),
                ),
                BuilderConfig(
                    name="testy-2",
                    workernames=["local"],
                    factory=BuildFactory(),
                ),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_ids = [
            (yield self.master.data.updates.findBuilderId('testy-1')),
            (yield self.master.data.updates.findBuilderId('testy-2')),
        ]

        return controller, builder_ids

    @defer.inlineCallbacks
    def reconfig_workers_remove_all(self):
        config_dict = {'workers': [], 'multiMaster': True}
        config = MasterConfig.loadFromDict(config_dict, '<dict>')
        yield self.master.workers.reconfigServiceWithBuildbotConfig(config)

    def stop_first_build(self, results):
        stopped_d = defer.Deferred()

        def new_callback(_, data):
            if stopped_d.called:
                return

            # Stop the build
            buildid = data['buildid']
            self.master.mq.produce(
                ('control', 'builds', str(buildid), 'stop'),
                {'reason': 'no reason', 'results': results},
            )
            stopped_d.callback(None)

        consumed_d = self.master.mq.startConsuming(new_callback, ('builds', None, 'new'))

        return consumed_d, stopped_d

    @defer.inlineCallbacks
    def test_latent_workers_start_in_parallel(self):
        """
        If there are two latent workers configured, and two build
        requests for them, both workers will start substantiating
        concurrently.
        """
        controllers = [
            LatentController(self, 'local1'),
            LatentController(self, 'local2'),
        ]
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy", workernames=["local1", "local2"], factory=BuildFactory()
                ),
            ],
            'workers': [controller.worker for controller in controllers],
            'protocols': {'null': {}},
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_id = yield self.master.data.updates.findBuilderId('testy')

        # Request two builds.
        for _ in range(2):
            yield self.create_build_request([builder_id])

        # Check that both workers were requested to start.
        self.assertEqual(controllers[0].starting, True)
        self.assertEqual(controllers[1].starting, True)
        for controller in controllers:
            yield controller.start_instance(True)
            yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_refused_substantiations_get_requeued(self):
        """
        If a latent worker refuses to substantiate, the build request becomes
        unclaimed.
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )

        # Indicate that the worker can't start an instance.
        yield controller.start_instance(False)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})

        yield self.assertBuildResults(1, RETRY)
        yield controller.auto_stop(True)
        self.flushLoggedErrors(LatentWorkerFailedToSubstantiate)

    @defer.inlineCallbacks
    def test_failed_substantiations_get_requeued(self):
        """
        If a latent worker fails to substantiate, the build request becomes
        unclaimed.
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )

        # The worker fails to substantiate.
        yield controller.start_instance(Failure(TestException("substantiation failed")))
        # Flush the errors logged by the failure.
        self.flushLoggedErrors(TestException)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        yield self.assertBuildResults(1, RETRY)
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_failed_substantiations_get_exception(self):
        """
        If a latent worker fails to substantiate, the result is an exception.
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        yield self.create_build_request([builder_id])

        # The worker fails to substantiate.
        yield controller.start_instance(
            Failure(LatentWorkerCannotSubstantiate("substantiation failed"))
        )
        # Flush the errors logged by the failure.
        self.flushLoggedErrors(LatentWorkerCannotSubstantiate)

        # When the substantiation fails, the result is an exception.
        yield self.assertBuildResults(1, EXCEPTION)
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_worker_accepts_builds_after_failure(self):
        """
        If a latent worker fails to substantiate, the worker is still able to
        accept jobs.
        """
        controller, builder_id = yield self.create_single_worker_config()

        yield controller.auto_stop(True)
        # Trigger a buildrequest
        yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )
        # The worker fails to substantiate.
        yield controller.start_instance(Failure(TestException("substantiation failed")))
        # Flush the errors logged by the failure.
        self.flushLoggedErrors(TestException)

        # The retry logic should only trigger after a exponential backoff
        self.assertEqual(controller.starting, False)

        # advance the time to the point where we should retry
        self.reactor.advance(controller.worker.quarantine_initial_timeout)

        # If the worker started again after the failure, then the retry logic will have
        # already kicked in to start a new build on this (the only) worker. We check that
        # a new instance was requested, which indicates that the worker
        # accepted the build.
        self.assertEqual(controller.starting, True)

        # The worker fails to substantiate(again).
        yield controller.start_instance(Failure(TestException("substantiation failed")))
        # Flush the errors logged by the failure.
        self.flushLoggedErrors(TestException)

        yield self.assertBuildResults(1, RETRY)

        # advance the time to the point where we should not retry
        self.reactor.advance(controller.worker.quarantine_initial_timeout)
        self.assertEqual(controller.starting, False)
        # advance the time to the point where we should retry
        self.reactor.advance(controller.worker.quarantine_initial_timeout)
        self.assertEqual(controller.starting, True)

        controller.auto_start(True)
        controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_worker_multiple_substantiations_succeed(self):
        """
        If multiple builders trigger try to substantiate a worker at
        the same time, if the substantiation succeeds then all of
        the builds proceed.
        """
        controller, builder_ids = yield self.create_single_worker_two_builder_config()

        # Trigger a buildrequest
        yield self.create_build_request(builder_ids)

        # The worker succeeds to substantiate.
        yield controller.start_instance(True)

        yield self.assertBuildResults(1, SUCCESS)
        yield self.assertBuildResults(2, SUCCESS)

        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_very_late_detached_after_substantiation(self):
        """
        A latent worker may detach at any time after stop_instance() call.
        Make sure it works at the most late detachment point, i.e. when we're
        substantiating again.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        yield self.create_build_request([builder_id])

        self.assertTrue(controller.starting)

        controller.auto_disconnect_worker = False
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        self.reactor.advance(1)

        # stop the instance, but don't disconnect the worker up to until just
        # before we complete start_instance()
        self.assertTrue(controller.stopping)
        yield controller.stop_instance(True)

        self.assertTrue(controller.stopped)
        yield self.create_build_request([builder_id])
        self.assertTrue(controller.starting)
        yield controller.disconnect_worker()
        yield controller.start_instance(True)

        yield self.assertBuildResults(2, SUCCESS)
        self.reactor.advance(1)

        yield controller.stop_instance(True)
        yield controller.disconnect_worker()

    @defer.inlineCallbacks
    def test_substantiation_during_stop_instance(self):
        """
        If a latent worker detaches before stop_instance() completes and we
        start a build then it should start successfully without causing an
        erroneous cancellation of the substantiation request.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        # Trigger a single buildrequest
        yield self.create_build_request([builder_id])

        self.assertEqual(True, controller.starting)

        # start instance
        controller.auto_disconnect_worker = False
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        self.reactor.advance(1)

        self.assertTrue(controller.stopping)
        yield controller.disconnect_worker()
        # now create a buildrequest that will substantiate the build. It should
        # either not start at all until the instance finished substantiating,
        # or the substantiation request needs to be recorded and start
        # immediately after stop_instance completes.
        yield self.create_build_request([builder_id])
        yield controller.stop_instance(True)

        yield controller.start_instance(True)
        yield self.assertBuildResults(2, SUCCESS)

        self.reactor.advance(1)
        yield controller.stop_instance(True)
        yield controller.disconnect_worker()

    @defer.inlineCallbacks
    def test_substantiation_during_stop_instance_canStartBuild_race(self):
        """
        If build attempts substantiation after the latent worker detaches,
        but stop_instance() is not completed yet, then we should successfully
        complete substantiation without causing an erroneous cancellation.
        The above sequence of events was possible even if canStartBuild
        checked for a in-progress insubstantiation, as if the build is scheduled
        before insubstantiation, its start could be delayed until when
        stop_instance() is in progress.
        """
        controller, builder_ids = yield self.create_single_worker_two_builder_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        # Trigger a single buildrequest
        yield self.create_build_request([builder_ids[0]])

        self.assertEqual(True, controller.starting)

        # start instance
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        with patchForDelay('buildbot.process.builder.Builder.maybeStartBuild') as delay:
            # create a build request which will result in a build, but it won't
            # attempt to substantiate until after stop_instance() is in progress
            yield self.create_build_request([builder_ids[1]])
            self.assertEqual(len(delay), 1)
            self.reactor.advance(1)

            self.assertTrue(controller.stopping)
            delay.fire()
            yield controller.stop_instance(True)

        self.assertTrue(controller.starting)
        yield controller.start_instance(True)
        yield self.assertBuildResults(2, SUCCESS)

        self.reactor.advance(1)
        yield controller.stop_instance(True)

    @defer.inlineCallbacks
    def test_insubstantiation_during_substantiation_refuses_substantiation(self):
        """
        If a latent worker gets insubstantiation() during substantiation, then it should refuse
        to substantiate.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        # insubstantiate during start_instance(). Note that failed substantiation is notified only
        # after the latent workers completes start-stop cycle.
        yield self.create_build_request([builder_id])
        d = controller.worker.insubstantiate()
        yield controller.start_instance(False)
        yield controller.stop_instance(True)
        yield d

        yield self.assertBuildResults(1, RETRY)

    @defer.inlineCallbacks
    def test_stopservice_during_insubstantiation_completes(self):
        """
        When stopService is called and a worker is insubstantiating, we should wait for this
        process to complete.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)

        yield self.assertBuildResults(1, SUCCESS)
        self.assertTrue(controller.started)

        # Wait until worker starts insubstantiation and then shutdown worker
        self.reactor.advance(1)
        self.assertTrue(controller.stopping)

        d = self.reconfig_workers_remove_all()
        self.assertFalse(d.called)
        yield controller.stop_instance(True)
        yield d

    @parameterized.expand([
        ('with_substantiation_failure', False, False),
        ('without_worker_connecting', True, False),
        ('with_worker_connecting', True, True),
    ])
    @defer.inlineCallbacks
    def test_stopservice_during_substantiation_completes(
        self, name, subst_success, worker_connects
    ):
        """
        When stopService is called and a worker is substantiating, we should wait for this
        process to complete.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )
        controller.auto_connect_worker = worker_connects

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        self.assertTrue(controller.starting)

        d = self.reconfig_workers_remove_all()
        self.assertFalse(d.called)

        yield controller.start_instance(subst_success)

        # we should stop the instance immediately after it substantiates regardless of the result
        self.assertTrue(controller.stopping)

        yield controller.stop_instance(True)
        yield d

    @defer.inlineCallbacks
    def test_substantiation_is_cancelled_by_build_stop(self):
        """
        Stopping a build during substantiation should cancel the substantiation itself.
        Otherwise we will be left with a substantiating worker without a corresponding build
        which means that master shutdown may not work correctly.
        """
        controller, builder_id = yield self.create_single_worker_config()
        controller.auto_connect_worker = False
        controller.auto_stop(True)

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)

        self.master.mq.produce(('control', 'builds', '1', 'stop'), {'reason': 'no reason'})

        self.reactor.advance(1)  # force build to actually execute the stop instruction

        self.assertTrue(controller.stopped)

    @parameterized.expand([
        ('after_start_instance_no_worker', False, False),
        ('after_start_instance_with_worker', True, False),
        ('before_start_instance_no_worker', False, True),
        ('before_start_instance_with_worker', True, True),
    ])
    @defer.inlineCallbacks
    def test_reconfigservice_during_substantiation_clean_shutdown_after(
        self, name, worker_connects, before_start_service
    ):
        """
        When stopService is called and a worker is substantiating, we should wait for this
        process to complete.
        """
        registered_workers = []

        def registration_updates(reg, worker_config, global_config):
            registered_workers.append((worker_config.workername, worker_config.password))

        self.patch(manager.WorkerRegistration, 'update', registration_updates)

        controller, builder_id = yield self.create_single_worker_config()
        controller.auto_connect_worker = worker_connects
        controller.auto_stop(True)

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        self.assertTrue(controller.starting)

        # change some unimportant property of the worker to force configuration
        self.master.config_loader.config_dict['workers'] = [
            ControllableLatentWorker('local', controller, max_builds=3)
        ]

        if before_start_service:
            yield self.reconfig_master()
            yield controller.start_instance(True)
        else:
            yield controller.start_instance(True)
            yield self.reconfig_master()

        yield self.clean_master_shutdown(quick=True)

        self.assertEqual(registered_workers, [('local', 'password_1'), ('local', 'password_1')])

    @defer.inlineCallbacks
    def test_substantiation_cancelled_by_insubstantiation_when_waiting_for_insubstantiation(self):
        """
        We should cancel substantiation if we insubstantiate when that substantiation is waiting
        on current insubstantiation to finish
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        yield self.create_build_request([builder_id])

        # put the worker into insubstantiation phase
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        self.reactor.advance(1)
        self.assertTrue(controller.stopping)

        # build should wait on the insubstantiation
        yield self.create_build_request([builder_id])
        self.assertEqual(controller.worker.state, States.INSUBSTANTIATING_SUBSTANTIATING)

        # build should be requeued if we insubstantiate.
        d = controller.worker.insubstantiate()
        yield controller.stop_instance(True)
        yield d

        yield self.assertBuildResults(2, RETRY)

    @defer.inlineCallbacks
    def test_stalled_substantiation_then_timeout_get_requeued(self):
        """
        If a latent worker substantiate, but not connect, and then be
        unsubstantiated, the build request becomes unclaimed.
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )

        # We never start the worker, rather timeout it.
        self.reactor.advance(controller.worker.missing_timeout)
        # Flush the errors logged by the failure.
        self.flushLoggedErrors(defer.TimeoutError)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        yield controller.start_instance(False)
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_stalled_substantiation_then_check_instance_fails_get_requeued(self):
        """
        If a latent worker substantiate, but not connect and check_instance() indicates a crash,
        the build request should become unclaimed as soon as check_instance_interval passes
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={'check_instance_interval': 10}
        )
        controller.auto_connect_worker = False

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )

        # The worker startup succeeds, but it doe not connect and check_instance() later
        # indicates a crash.
        yield controller.start_instance(True)

        self.reactor.advance(10)
        controller.has_crashed = True
        self.reactor.advance(10)

        # Flush the errors logged by the failure.
        self.flushLoggedErrors(LatentWorkerFailedToSubstantiate)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_sever_connection_before_ping_then_timeout_get_requeued(self):
        """
        If a latent worker connects, but its connection is severed without
        notification in the TCP layer, we successfully wait until TCP times
        out and requeue the build.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        yield self.create_build_request([builder_id])

        # sever connection just before ping()
        with patchForDelay(
            'buildbot.process.workerforbuilder.AbstractWorkerForBuilder.ping'
        ) as delay:
            yield controller.start_instance(True)
            controller.sever_connection()
            delay.fire()

        # lose connection after TCP times out
        self.reactor.advance(100)
        yield controller.disconnect_worker()
        yield self.assertBuildResults(1, RETRY)

        # the worker will be put into quarantine
        self.reactor.advance(controller.worker.quarantine_initial_timeout)

        yield controller.stop_instance(True)
        yield controller.start_instance(True)
        yield self.assertBuildResults(2, SUCCESS)

        self.reactor.advance(1)
        yield controller.stop_instance(True)

        self.flushLoggedErrors(pb.PBConnectionLost)

    @defer.inlineCallbacks
    def test_failed_sendBuilderList_get_requeued(self):
        """
        sendBuilderList can fail due to missing permissions on the workdir,
        the build request becomes unclaimed
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )
        logs = []
        yield self.master.mq.startConsuming(
            lambda key, log: logs.append(log), ('logs', None, 'new')
        )

        # The worker succeed to substantiate
        def remote_setBuilderList(self, dirs):
            raise TestException("can't create dir")

        controller.patchBot(self, 'remote_setBuilderList', remote_setBuilderList)
        yield controller.start_instance(True)

        # Flush the errors logged by the failure.
        self.flushLoggedErrors(TestException)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        # should get 2 logs (html and txt) with proper information in there
        self.assertEqual(len(logs), 2)
        logs_by_name = {}
        for _log in logs:
            fulllog = yield self.master.data.get(("logs", str(_log['logid']), "raw"))
            logs_by_name[fulllog['filename']] = fulllog['raw']

        for i in [
            'testy_build_1_step_worker_preparation_log_err_text',
            "testy_build_1_step_worker_preparation_log_err_html",
        ]:
            self.assertIn("can't create dir", logs_by_name[i])
            # make sure stacktrace is present in html
            self.assertIn(
                "buildbot.test.integration.test_worker_latent.TestException", logs_by_name[i]
            )
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_failed_ping_get_requeued(self):
        """
        sendBuilderList can fail due to missing permissions on the workdir,
        the build request becomes unclaimed
        """
        controller, builder_id = yield self.create_single_worker_config()

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )
        logs = []
        yield self.master.mq.startConsuming(
            lambda key, log: logs.append(log), ('logs', None, 'new')
        )

        # The worker succeed to substantiate
        def remote_print(self, msg):
            if msg == "ping":
                raise TestException("can't ping")

        controller.patchBot(self, 'remote_print', remote_print)
        yield controller.start_instance(True)

        # Flush the errors logged by the failure.
        self.flushLoggedErrors(TestException)

        # When the substantiation fails, the buildrequest becomes unclaimed.
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        # should get 2 logs (html and txt) with proper information in there
        self.assertEqual(len(logs), 2)
        logs_by_name = {}
        for _log in logs:
            fulllog = yield self.master.data.get(("logs", str(_log['logid']), "raw"))
            logs_by_name[fulllog['filename']] = fulllog['raw']

        for i in [
            'testy_build_1_step_worker_preparation_log_err_text',
            "testy_build_1_step_worker_preparation_log_err_html",
        ]:
            self.assertIn("can't ping", logs_by_name[i])
            # make sure stacktrace is present in html
            self.assertIn(
                "buildbot.test.integration.test_worker_latent.TestException", logs_by_name[i]
            )
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_worker_close_connection_while_building(self):
        """
        If the worker close connection in the middle of the build, the next
        build can start correctly
        """
        controller, stepcontroller, builder_id = yield self.create_single_worker_config_with_step(
            controller_kwargs={"build_wait_timeout": 0}
        )

        # Request a build and disconnect midway
        controller.auto_disconnect_worker = False
        yield self.create_build_request([builder_id])
        yield controller.auto_stop(True)

        self.assertTrue(controller.starting)
        yield controller.start_instance(True)

        yield self.assertBuildResults(1, None)
        yield controller.disconnect_worker()
        yield self.assertBuildResults(1, RETRY)

        # Now check that the build requeued and finished with success
        # after the debounce delay
        self.reactor.advance(1)
        yield controller.start_instance(True)

        yield self.assertBuildResults(2, None)
        stepcontroller.finish_step(SUCCESS)
        yield self.assertBuildResults(2, SUCCESS)
        yield controller.disconnect_worker()

    @defer.inlineCallbacks
    def test_negative_build_timeout_reattach_substantiated(self):
        """
        When build_wait_timeout is negative, we don't disconnect the worker from
        our side. We should still support accidental disconnections from
        worker side due to, e.g. network problems.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": -1}
        )

        controller.auto_disconnect_worker = False
        controller.auto_connect_worker = False

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield controller.connect_worker()

        yield self.assertBuildResults(1, SUCCESS)
        self.assertTrue(controller.started)

        # Now disconnect and reconnect worker and check whether we can still
        # build. This should not change the worker state from our side.
        yield controller.disconnect_worker()
        self.assertTrue(controller.started)
        yield controller.connect_worker()
        self.assertTrue(controller.started)

        yield self.create_build_request([builder_id])
        yield self.assertBuildResults(1, SUCCESS)

        # The only way to stop worker with negative build timeout is to
        # insubstantiate explicitly
        yield controller.auto_stop(True)
        yield controller.worker.insubstantiate()
        yield controller.disconnect_worker()

    @defer.inlineCallbacks
    def test_sever_connection_while_building(self):
        """
        If the connection to worker is severed without TCP notification in the
        middle of the build, the build is re-queued and successfully restarted.
        """
        controller, stepcontroller, builder_id = yield self.create_single_worker_config_with_step(
            controller_kwargs={"build_wait_timeout": 0}
        )

        # Request a build and disconnect midway
        yield self.create_build_request([builder_id])
        yield controller.auto_stop(True)

        self.assertTrue(controller.starting)
        yield controller.start_instance(True)

        yield self.assertBuildResults(1, None)
        # sever connection and lose it after TCP times out
        controller.sever_connection()
        self.reactor.advance(100)
        yield controller.disconnect_worker()

        yield self.assertBuildResults(1, RETRY)

        # Request one build.
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield self.assertBuildResults(2, None)
        stepcontroller.finish_step(SUCCESS)
        yield self.assertBuildResults(2, SUCCESS)

    @defer.inlineCallbacks
    def test_sever_connection_during_insubstantiation(self):
        """
        If latent worker connection is severed without notification in the TCP
        layer, we successfully wait until TCP times out, insubstantiate and
        can substantiate after that.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        yield self.create_build_request([builder_id])

        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        # sever connection just before insubstantiation and lose it after TCP
        # times out
        with patchForDelay('buildbot.worker.base.AbstractWorker.disconnect') as delay:
            self.reactor.advance(1)
            self.assertTrue(controller.stopping)
            controller.sever_connection()
            delay.fire()

        yield controller.stop_instance(True)
        self.reactor.advance(100)
        yield controller.disconnect_worker()

        # create new build request and verify it works
        yield self.create_build_request([builder_id])

        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)
        self.reactor.advance(1)
        yield controller.stop_instance(True)

        self.flushLoggedErrors(pb.PBConnectionLost)

    @defer.inlineCallbacks
    def test_sever_connection_during_insubstantiation_and_buildrequest(self):
        """
        If latent worker connection is severed without notification in the TCP
        layer, we successfully wait until TCP times out, insubstantiate and
        can substantiate after that. In this the subsequent build request is
        created during insubstantiation
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        yield self.create_build_request([builder_id])

        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        # sever connection just before insubstantiation and lose it after TCP
        # times out
        with patchForDelay('buildbot.worker.base.AbstractWorker.disconnect') as delay:
            self.reactor.advance(1)
            self.assertTrue(controller.stopping)
            yield self.create_build_request([builder_id])
            controller.sever_connection()
            delay.fire()

        yield controller.stop_instance(True)
        self.reactor.advance(100)
        yield controller.disconnect_worker()

        # verify the previously created build successfully completes
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)
        self.reactor.advance(1)
        yield controller.stop_instance(True)

        self.flushLoggedErrors(pb.PBConnectionLost)

    @defer.inlineCallbacks
    def test_negative_build_timeout_reattach_insubstantiating(self):
        """
        When build_wait_timeout is negative, we don't disconnect the worker from
        our side, but it can disconnect and reattach from worker side due to,
        e.g. network problems.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": -1}
        )

        controller.auto_disconnect_worker = False
        controller.auto_connect_worker = False

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield controller.connect_worker()

        yield self.assertBuildResults(1, SUCCESS)
        self.assertTrue(controller.started)

        # Now start insubstantiation and disconnect and reconnect the worker.
        # It should not change worker state from master side.
        d = controller.worker.insubstantiate()
        self.assertTrue(controller.stopping)
        yield controller.disconnect_worker()
        self.assertTrue(controller.stopping)
        yield controller.connect_worker()
        self.assertTrue(controller.stopping)
        yield controller.stop_instance(True)
        yield d

        self.assertTrue(controller.stopped)
        yield controller.disconnect_worker()

        # Now substantiate the worker and verify build succeeds
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield controller.connect_worker()
        yield self.assertBuildResults(1, SUCCESS)

        controller.auto_disconnect_worker = True
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_negative_build_timeout_no_disconnect_insubstantiating(self):
        """
        When build_wait_timeout is negative, we don't disconnect the worker from
        our side, so it should be possible to insubstantiate and substantiate
        it without problems if the worker does not disconnect either.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": -1}
        )

        controller.auto_disconnect_worker = False
        controller.auto_connect_worker = False

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield controller.connect_worker()

        yield self.assertBuildResults(1, SUCCESS)
        self.assertTrue(controller.started)

        # Insubstantiate worker without disconnecting it
        d = controller.worker.insubstantiate()
        self.assertTrue(controller.stopping)
        yield controller.stop_instance(True)
        yield d

        self.assertTrue(controller.stopped)

        # Now substantiate the worker without connecting it
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        yield self.assertBuildResults(1, SUCCESS)

        controller.auto_disconnect_worker = True
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_negative_build_timeout_insubstantiates_on_master_shutdown(self):
        """
        When build_wait_timeout is negative, we should still insubstantiate when master shuts down.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": -1}
        )

        # Substantiate worker via a build
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)

        yield self.assertBuildResults(1, SUCCESS)
        self.assertTrue(controller.started)

        # Shutdown master
        d = self.master.stopService()
        yield controller.stop_instance(True)
        yield d

    @defer.inlineCallbacks
    def test_stop_instance_synchronous_exception(self):
        """
        Throwing a synchronous exception from stop_instance should allow subsequent build to start.
        """
        controller, builder_id = yield self.create_single_worker_config(
            controller_kwargs={"build_wait_timeout": 1}
        )

        controller.auto_stop(True)

        # patch stop_instance() to raise exception synchronously
        def raise_stop_instance(fast):
            raise TestException()

        real_stop_instance = controller.worker.stop_instance
        controller.worker.stop_instance = raise_stop_instance

        # create a build and wait for stop
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        self.reactor.advance(1)
        yield self.assertBuildResults(1, SUCCESS)
        self.flushLoggedErrors(TestException)

        # unpatch stop_instance() and call it to cleanup state of fake worker controller
        controller.worker.stop_instance = real_stop_instance
        yield controller.worker.stop_instance(False)

        self.reactor.advance(1)

        # subsequent build should succeed
        yield self.create_build_request([builder_id])
        yield controller.start_instance(True)
        self.reactor.advance(1)
        yield self.assertBuildResults(2, SUCCESS)

    @defer.inlineCallbacks
    def test_build_stop_with_cancelled_during_substantiation(self):
        """
        If a build is stopping during latent worker substantiating, the build
        becomes cancelled
        """
        controller, builder_id = yield self.create_single_worker_config()

        consumed_d, stopped_d = self.stop_first_build(CANCELLED)
        yield consumed_d

        # Trigger a buildrequest
        yield self.create_build_request([builder_id])

        yield stopped_d

        # Indicate that the worker can't start an instance.
        yield controller.start_instance(False)

        yield self.assertBuildResults(1, CANCELLED)

        yield controller.auto_stop(True)
        self.flushLoggedErrors(LatentWorkerFailedToSubstantiate)

    @defer.inlineCallbacks
    def test_build_stop_with_retry_during_substantiation(self):
        """
        If master is shutting down during latent worker substantiating, the build becomes retry.
        """
        controller, builder_id = yield self.create_single_worker_config()

        consumed_d, stopped_d = self.stop_first_build(RETRY)
        yield consumed_d

        # Trigger a buildrequest
        _, brids = yield self.create_build_request([builder_id])

        unclaimed_build_requests = []
        yield self.master.mq.startConsuming(
            lambda key, request: unclaimed_build_requests.append(request),
            ('buildrequests', None, 'unclaimed'),
        )

        yield stopped_d

        # Indicate that the worker can't start an instance.
        yield controller.start_instance(False)
        self.reactor.advance(1)

        yield self.assertBuildResults(1, RETRY)
        self.assertEqual(set(brids), {req['buildrequestid'] for req in unclaimed_build_requests})
        yield controller.auto_stop(True)
        yield controller.start_instance(False)
        self.flushLoggedErrors(LatentWorkerFailedToSubstantiate)

    @defer.inlineCallbacks
    def test_rejects_build_on_instance_with_different_type_timeout_zero(self):
        """
        If latent worker supports getting its instance type from properties that
        are rendered from build then the buildrequestdistributor must not
        schedule any builds on workers that are running different instance type
        than what these builds will require.
        """
        controller, stepcontroller, builder_id = yield self.create_single_worker_config_with_step(
            controller_kwargs={"kind": Interpolate('%(prop:worker_kind)s'), "build_wait_timeout": 0}
        )

        # create build request
        yield self.create_build_request([builder_id], properties=Properties(worker_kind='a'))

        # start the build and verify the kind of the worker. Note that the
        # buildmaster needs to restart the worker in order to change the worker
        # kind, so we allow it both to auto start and stop
        self.assertEqual(True, controller.starting)

        controller.auto_start(True)
        yield controller.auto_stop(True)
        self.assertEqual((yield controller.get_started_kind()), 'a')

        # before the other build finished, create another build request
        yield self.create_build_request([builder_id], properties=Properties(worker_kind='b'))
        stepcontroller.finish_step(SUCCESS)

        # give the botmaster chance to insubstantiate the worker and
        # maybe substantiate it for the pending build the builds on worker
        self.reactor.advance(0.1)

        # verify that the second build restarted with the expected instance
        # kind
        self.assertEqual((yield controller.get_started_kind()), 'b')
        stepcontroller.finish_step(SUCCESS)

        yield self.assertBuildResults(1, SUCCESS)
        yield self.assertBuildResults(2, SUCCESS)

    @defer.inlineCallbacks
    def test_rejects_build_on_instance_with_different_type_timeout_nonzero(self):
        """
        If latent worker supports getting its instance type from properties that
        are rendered from build then the buildrequestdistributor must not
        schedule any builds on workers that are running different instance type
        than what these builds will require.
        """

        controller, stepcontroller, builder_id = yield self.create_single_worker_config_with_step(
            controller_kwargs={"kind": Interpolate('%(prop:worker_kind)s'), "build_wait_timeout": 5}
        )

        # create build request
        yield self.create_build_request([builder_id], properties=Properties(worker_kind='a'))

        # start the build and verify the kind of the worker. Note that the
        # buildmaster needs to restart the worker in order to change the worker
        # kind, so we allow it both to auto start and stop
        self.assertEqual(True, controller.starting)

        controller.auto_start(True)
        yield controller.auto_stop(True)
        self.assertEqual((yield controller.get_started_kind()), 'a')

        # before the other build finished, create another build request
        yield self.create_build_request([builder_id], properties=Properties(worker_kind='b'))
        stepcontroller.finish_step(SUCCESS)

        # give the botmaster chance to insubstantiate the worker and
        # maybe substantiate it for the pending build the builds on worker
        self.reactor.advance(0.1)

        # verify build has not started, even though the worker is waiting
        # for one
        self.assertIsNone((yield self.master.db.builds.getBuild(2)))
        self.assertTrue(controller.started)

        # wait until the latent worker times out, is insubstantiated,
        # is substantiated because of pending buildrequest and starts the build
        self.reactor.advance(6)
        self.assertIsNotNone((yield self.master.db.builds.getBuild(2)))

        # verify that the second build restarted with the expected instance
        # kind
        self.assertEqual((yield controller.get_started_kind()), 'b')
        stepcontroller.finish_step(SUCCESS)

        yield self.assertBuildResults(1, SUCCESS)
        yield self.assertBuildResults(2, SUCCESS)

    @defer.inlineCallbacks
    def test_supports_no_build_for_substantiation(self):
        """
        Abstract latent worker should support being substantiated without a
        build and then insubstantiated.
        """
        controller, _ = yield self.create_single_worker_config()

        d = controller.worker.substantiate(None, None)
        yield controller.start_instance(True)
        self.assertTrue(controller.started)
        yield d

        d = controller.worker.insubstantiate()
        yield controller.stop_instance(True)
        yield d

    @defer.inlineCallbacks
    def test_supports_no_build_for_substantiation_accepts_build_later(self):
        """
        Abstract latent worker should support being substantiated without a
        build and then accept a build request.
        """
        controller, stepcontroller, builder_id = yield self.create_single_worker_config_with_step(
            controller_kwargs={"build_wait_timeout": 1}
        )

        d = controller.worker.substantiate(None, None)
        yield controller.start_instance(True)
        self.assertTrue(controller.started)
        yield d

        self.create_build_request([builder_id])
        stepcontroller.finish_step(SUCCESS)

        self.reactor.advance(1)
        yield controller.stop_instance(True)


class LatentWithLatentMachine(TimeoutableTestCase, RunFakeMasterTestCase):
    def tearDown(self):
        # Flush the errors logged by the master stop cancelling the builds.
        self.flushLoggedErrors(LatentWorkerSubstantiatiationCancelled)
        super().tearDown()

    @defer.inlineCallbacks
    def create_single_worker_config(self, build_wait_timeout=0):
        machine_controller = LatentMachineController(
            name='machine1', build_wait_timeout=build_wait_timeout
        )

        worker_controller = LatentController(self, 'worker1', machine_name='machine1')
        step_controller = BuildStepController()

        config_dict = {
            'machines': [machine_controller.machine],
            'builders': [
                BuilderConfig(
                    name="builder1",
                    workernames=["worker1"],
                    factory=BuildFactory([step_controller.step]),
                ),
            ],
            'workers': [worker_controller.worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }

        yield self.setup_master(config_dict)
        builder_id = yield self.master.data.updates.findBuilderId('builder1')

        return machine_controller, worker_controller, step_controller, builder_id

    @defer.inlineCallbacks
    def create_two_worker_config(self, build_wait_timeout=0, controller_kwargs=None):
        if not controller_kwargs:
            controller_kwargs = {}

        machine_controller = LatentMachineController(
            name='machine1', build_wait_timeout=build_wait_timeout
        )

        worker1_controller = LatentController(
            self, 'worker1', machine_name='machine1', **controller_kwargs
        )
        worker2_controller = LatentController(
            self, 'worker2', machine_name='machine1', **controller_kwargs
        )
        step1_controller = BuildStepController()
        step2_controller = BuildStepController()

        config_dict = {
            'machines': [machine_controller.machine],
            'builders': [
                BuilderConfig(
                    name="builder1",
                    workernames=["worker1"],
                    factory=BuildFactory([step1_controller.step]),
                ),
                BuilderConfig(
                    name="builder2",
                    workernames=["worker2"],
                    factory=BuildFactory([step2_controller.step]),
                ),
            ],
            'workers': [worker1_controller.worker, worker2_controller.worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }

        yield self.setup_master(config_dict)
        builder1_id = yield self.master.data.updates.findBuilderId('builder1')
        builder2_id = yield self.master.data.updates.findBuilderId('builder2')

        return (
            machine_controller,
            [worker1_controller, worker2_controller],
            [step1_controller, step2_controller],
            [builder1_id, builder2_id],
        )

    @defer.inlineCallbacks
    def test_1worker_starts_and_stops_after_single_build_success(self):
        (
            machine_controller,
            worker_controller,
            step_controller,
            builder_id,
        ) = yield self.create_single_worker_config()

        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])
        machine_controller.start_machine(True)
        self.assertTrue(worker_controller.started)

        step_controller.finish_step(SUCCESS)
        self.reactor.advance(0)  # force deferred suspend call to be executed
        machine_controller.stop_machine()

        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_1worker_starts_and_stops_after_single_build_failure(self):
        (
            machine_controller,
            worker_controller,
            step_controller,
            builder_id,
        ) = yield self.create_single_worker_config()

        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])
        machine_controller.start_machine(True)
        self.assertTrue(worker_controller.started)

        step_controller.finish_step(FAILURE)
        self.reactor.advance(0)  # force deferred stop call to be executed
        machine_controller.stop_machine()

        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_1worker_stops_machine_after_timeout(self):
        (
            machine_controller,
            worker_controller,
            step_controller,
            builder_id,
        ) = yield self.create_single_worker_config(build_wait_timeout=5)

        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])

        machine_controller.start_machine(True)
        self.reactor.advance(10.0)
        step_controller.finish_step(SUCCESS)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)

        self.reactor.advance(4.9)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)

        # put clock 5s after step finish, machine should start suspending
        self.reactor.advance(0.1)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPING)

        machine_controller.stop_machine()

        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_1worker_does_not_stop_machine_machine_after_timeout_during_build(self):
        (
            machine_controller,
            worker_controller,
            step_controller,
            builder_id,
        ) = yield self.create_single_worker_config(build_wait_timeout=5)

        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])

        machine_controller.start_machine(True)
        self.reactor.advance(10.0)
        step_controller.finish_step(SUCCESS)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)

        # create build request while machine is still awake. It should not
        # suspend regardless of how much time passes

        self.reactor.advance(3.9)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)
        yield self.create_build_request([builder_id])

        self.reactor.advance(5.1)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)

        step_controller.finish_step(SUCCESS)
        self.reactor.advance(4.9)
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTED)

        # put clock 5s after step finish, machine should start suspending
        self.reactor.advance(0.1)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPING)

        machine_controller.stop_machine()

        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_1worker_insubstantiated_after_start_failure(self):
        (
            machine_controller,
            worker_controller,
            _,
            builder_id,
        ) = yield self.create_single_worker_config()

        worker_controller.auto_connect_worker = False
        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])

        machine_controller.start_machine(False)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)
        self.assertEqual(worker_controller.started, False)

    @defer.inlineCallbacks
    def test_1worker_eats_exception_from_start_machine(self):
        (
            machine_controller,
            worker_controller,
            _,
            builder_id,
        ) = yield self.create_single_worker_config()

        worker_controller.auto_connect_worker = False
        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])

        class FakeError(Exception):
            pass

        machine_controller.start_machine(FakeError('start error'))
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)
        self.assertEqual(worker_controller.started, False)

        self.flushLoggedErrors(FakeError)

    @defer.inlineCallbacks
    def test_1worker_eats_exception_from_stop_machine(self):
        (
            machine_controller,
            worker_controller,
            step_controller,
            builder_id,
        ) = yield self.create_single_worker_config()

        worker_controller.auto_start(True)
        worker_controller.auto_stop(True)

        yield self.create_build_request([builder_id])

        machine_controller.start_machine(True)
        step_controller.finish_step(SUCCESS)
        self.reactor.advance(0)  # force deferred suspend call to be executed

        class FakeError(Exception):
            pass

        machine_controller.stop_machine(FakeError('stop error'))

        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)
        self.flushLoggedErrors(FakeError)

    @defer.inlineCallbacks
    def test_2workers_build_substantiates_insubstantiates_both_workers(self):
        (
            machine_controller,
            worker_controllers,
            step_controllers,
            builder_ids,
        ) = yield self.create_two_worker_config(
            controller_kwargs={"starts_without_substantiate": True}
        )

        for wc in worker_controllers:
            wc.auto_start(True)
            wc.auto_stop(True)

        yield self.create_build_request([builder_ids[0]])

        machine_controller.start_machine(True)
        for wc in worker_controllers:
            self.assertTrue(wc.started)

        step_controllers[0].finish_step(SUCCESS)
        self.reactor.advance(0)  # force deferred suspend call to be executed
        machine_controller.stop_machine()

        for wc in worker_controllers:
            self.assertFalse(wc.started)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_2workers_two_builds_start_machine_concurrently(self):
        (
            machine_controller,
            worker_controllers,
            step_controllers,
            builder_ids,
        ) = yield self.create_two_worker_config()

        for wc in worker_controllers:
            wc.auto_start(True)
            wc.auto_stop(True)

        yield self.create_build_request([builder_ids[0]])
        self.assertEqual(machine_controller.machine.state, MachineStates.STARTING)

        yield self.create_build_request([builder_ids[1]])

        machine_controller.start_machine(True)
        for wc in worker_controllers:
            self.assertTrue(wc.started)

        step_controllers[0].finish_step(SUCCESS)
        step_controllers[1].finish_step(SUCCESS)
        self.reactor.advance(0)  # force deferred suspend call to be executed
        machine_controller.stop_machine()

        for wc in worker_controllers:
            self.assertFalse(wc.started)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

    @defer.inlineCallbacks
    def test_2workers_insubstantiated_after_one_start_failure(self):
        (
            machine_controller,
            worker_controllers,
            _,
            builder_ids,
        ) = yield self.create_two_worker_config()

        for wc in worker_controllers:
            wc.auto_connect_worker = False
            wc.auto_start(True)
            wc.auto_stop(True)

        yield self.create_build_request([builder_ids[0]])

        machine_controller.start_machine(False)
        self.assertEqual(machine_controller.machine.state, MachineStates.STOPPED)

        for wc in worker_controllers:
            self.assertEqual(wc.started, False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""add pause_reason column to workers table

Revision ID: 066
Revises: 065

"""

import sqlalchemy as sa
from alembic import op

from buildbot.util import sautils

# revision identifiers, used by Alembic.
revision = "066"
down_revision = "065"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("builds", sa.Column("locks_duration_s", sa.Integer, nullable=True))

    metadata = sa.MetaData()
    builds_tbl = sautils.Table(
        "builds", metadata, sa.Column("locks_duration_s", sa.Integer, nullable=True)
    )

    op.execute(builds_tbl.update().values({builds_tbl.c.locks_duration_s: 0}))

    with op.batch_alter_table("builds") as batch_op:
        batch_op.alter_column("locks_duration_s", existing_type=sa.Integer, nullable=False)


def downgrade():
    op.drop_column("builds", "locks_duration_s")
